from mpi4py import MPI
import numpy as np
import sys,os,time
import healpy as HP
from pylab import *
import cosmology_library as CL
import scipy.fftpack
import scipy.integrate as si
import scipy.special as ss
from scipy.optimize import minimize
import emcee, corner
import math
import gc

###### MPI DEFINITIONS ######                                         
comm   = MPI.COMM_WORLD
nprocs = comm.Get_size()
myrank = comm.Get_rank()

###############################################################################
# These functions return the value of Omega_HI(z),b_HI and b_21cm for the model
def Omega_HI_model(z):
    return 4e-4*(1.0+z)**0.6

def HI_bias_model(z):
    return 0.904 + 0.135*(1.0+z)**1.696

def bias_21cm_model(z,Omega_m,Omega_L,h):
    Dz      = CL.linear_growth_factor(z,Omega_m,Omega_L)[0]
    HI_bias = HI_bias_model(z)
    mean_T  = 190.0*(1.0+z)**2/np.sqrt(Omega_m*(1.0+z)**3+Omega_L)*\
        Omega_HI_model(z)*h #mK
    b_21cm = Dz*mean_T*HI_bias
    return b_21cm
###############################################################################

###############################################################################
# This routine returns the dictionary survey that contains the parameters of
# the survey such as mean redshift, BoxSize, dims...etc
# bins_min,bins_max --------> consider maps between [bins_min,bins_max]
# nuTable ------------------> file with frequencies and redshifts of maps
# cosmo --------------------> dictionary with cosmological parameters
# instrument ---------------> dictionary with instrument parameters
def get_survey_params(bin_min,bin_max,nuTable,cosmo,instrument):

    # compute value of dims for the 
    dims = bin_max - bin_min + 1

    # read the frequencies and redshifts of the maps
    num, nui, nuf, zi, zf = np.loadtxt(nuTable,unpack=True)

    # compute the comoving distance to the different maps
    d_co = np.zeros(len(num),dtype=np.float32);  z = 0.5*(zi+zf)
    for i in range(len(z)):  
        d_co[i] = CL.comoving_distance(z[i],cosmo['Omega_m'],
                                       cosmo['Omega_L'])  #Mpc/h

    # for the maps selected compute BoxSize and mean distance among maps
    # notice that maps begin with 1, while z and d_co array begin with 0
    BoxSize = d_co[bin_min-1]-d_co[bin_max-1]  #Mpc/h
    print('pixel separation = %.3f Mpc/h'%(BoxSize/dims))
    print('BoxSize =',BoxSize,'Mpc/h')

    # compute maps mean frequency, Bandwidth, wavelength and angular resolution
    nu_map     = 0.5*(nui+nuf)              #MHz
    Bnu_map    = nuf-nui                    #MHz
    lambda_map = 3e8/(nu_map*1e6)           #meters
    beam_fwhm  = lambda_map/instrument['D'] #radians

    # compute the mean redshift of the survey
    mean_z = 0.5*(z[bin_min-1]+z[bin_max-1]);  print('survey <z> = %.3f'%mean_z)

    survey = {'dims':dims,       'BoxSize':BoxSize,       'mean_z':mean_z, 
              'Bnu_map':Bnu_map, 'lambda_map':lambda_map, 'beam_fwhm':beam_fwhm,
              'z_map':z,         'nu_map':nu_map}

    return survey
###############################################################################

###############################################################################
# This routine computes the expected noise level and variance from the 
# instrument temperature
# cosmo ------------------------> dictionary with cosmological parameters
# survey -----------------------> dictionary with survey parameters
# instrument -------------------> dictionary with instrument parameters
# z ----------------------------> mean redshift of the maps
# Tsystem ----------------------> system temperature at maps mean redshift
def noise_level_variance(cosmo,survey,instrument,z,Tsystem):

    # value of H(z) in km/s/(Mpc/h)
    Hz = 100.0*np.sqrt(cosmo['Omega_m']*(1.0+z)**3 + cosmo['Omega_L'])
    
    # mean brightness temperature at z in mK
    Tb = 190.0*(100.0*(1.0+z)**2/Hz)*Omega_HI_model(z)*cosmo['h'] 

    # compute noise level and variance in noise for auto-correlations
    noise_level = 0.5*0.21e-3*(1.0+z)**2/Hz*(Tsystem/Tb)**2/\
        (survey['t_pixel']*instrument['n_dish'])  #Mpc/h
    sigma_noise = noise_level/np.sqrt(survey['pixels'])
    survey['noise_level_auto'] = noise_level
    survey['sigma_noise_auto'] = sigma_noise

    # correct above numbers for cross-correlations. Total time is the same for
    # auto and cross-correlations
    sigma_noise /= np.sqrt(2.0)  #Mpc/h 
    noise_level = 0.0            #Mpc/h
    survey['noise_level_cross'] = noise_level
    survey['sigma_noise_cross'] = sigma_noise
###############################################################################

###############################################################################
# This routine computes the value of sigma_T for the 21cm IM pixels
# It returns arrays with the system temperature and sigma_T for the different 
# maps and the time per pixel
# cosmo ------------------> dictionary with cosmological parameters
# survey -----------------> dictionary with survey parameters
# instrument -------------> dictionary with instrument parameters
def sigmaT(survey,instrument):

    T_instrument = instrument['T_instrument']  #K
    n_dish       = instrument['n_dish']

    # compute the system temperature 
    T_sky    = 60*(survey['nu_map']/300.)**(-2.5) #Atmosphere temperature
    T_system = (T_instrument + T_sky)*1e3  #Total system temperature in mK

    # compute sigma_T. pixels is total number of pixels in whole sky. If survey
    # is covering a fraction fsky, the number of pixels surveyed is fsky*pixels
    t_pixel = survey['time_total']*3600.0/(survey['fsky']*survey['pixels']) #sec
    sigma_T = T_system/np.sqrt(2.0*survey['Bnu_map']*1e6*t_pixel*n_dish) #mK

    survey['T_system'] = T_system
    survey['t_pixel']  = t_pixel
    survey['sigma_T']  = sigma_T
###############################################################################

###############################################################################
# This routine computes the radial P(k) and xi(r) in each patch of the sky.
# The number of patchs on the sky is given by pixels/subpixels. This routine
# can be used to compute the mean P(k) and xi(r) in the whole sky and to 
# estimate the JK errors
# cosmo ------------------> dictionary with cosmological parameters
# survey -----------------> dictionary with survey parameters
# instrument -------------> dictionary with instrument parameters
# delta_k2 ---------------> structure with values of delta^2(k) for all los
def clustering_region(cosmo,survey,instrument,delta_k2,fmask,do_JK):
    
    # read mask
    mask = HP.read_map(fmask)

    dims = survey['dims'];  BoxSize   = survey['BoxSize']
    #subpixels = survey['subpixels'];  

    # compute the values of k and r of the modes for the 1D P(k) and xi(r)
    modes   = np.arange(dims,dtype=np.float64);  middle = dims//2
    indexes = np.where(modes>middle)[0];  modes[indexes] = modes[indexes]-dims
    k = modes*(2.0*np.pi/BoxSize)  #k in h/Mpc
    r = modes*(BoxSize*1.0/dims)   #r in Mpc/h
    k = np.absolute(k);  r = np.absolute(r)  #just take the modulus
    del indexes, modes

    # define the k-bins and r-bins
    k_bins = np.linspace(0,dims//2,dims//2+1)*(2.0*np.pi/BoxSize)
    r_bins = np.linspace(0,dims//2,dims//2+1)*(BoxSize*1.0/dims)

    # compute the number of modes and the average number-weighted value of k,r
    k_modes = np.histogram(k,bins=k_bins)[0]
    k_bin   = np.histogram(k,bins=k_bins,weights=k)[0]/k_modes
    r_modes = np.histogram(r,bins=r_bins)[0]
    r_bin   = np.histogram(r,bins=r_bins,weights=r)[0]/r_modes


    # compute P(k) over the lines of sight not masked out
    if do_JK==False:

        # find the lines of sight over which compute the radial P(k)
        indexes_los = np.where(mask==1.0)[0]
        
        ############################## Pk(k) ###############################
        # find the delta_k2 matrix of the region
        delta_k2_region = delta_k2[:,indexes_los]

        # take all LOS and compute the average value for each mode
        delta_k2_stacked = np.mean(delta_k2_region,dtype=np.float64,axis=1)

        # compute the 1D P(k)
        Pk_mean = np.histogram(k,bins=k_bins,weights=delta_k2_stacked)[0]
        Pk_mean = Pk_mean/(BoxSize*k_modes);  del delta_k2_stacked

        ############################## xi(r) ###############################
        # Fourier transform P(k) to obtain xi(r)
        xi = scipy.fftpack.fftn(delta_k2_region,overwrite_x=True,axes=(0,))
        del delta_k2_region

        # take all LOS and compute the average value for each mode
        xi_stacked = np.mean(xi,dtype=np.float64,axis=1);  del xi

        # compute the 1D xi(r)
        xi_mean = np.histogram(r,bins=r_bins,weights=xi_stacked)[0]
        xi_mean = xi_mean/r_modes;  del xi_stacked, indexes_los
        
        # no JK errors are computed here
        Pk_var = np.zeros(len(Pk_mean),dtype=np.float32)
        xi_var = np.zeros(len(xi_mean),dtype=np.float32)
        
        return [k_bin,r_bin,Pk_mean,xi_mean,Pk_var,xi_var]


    # compute the number of maps to estimate variance with jackknife
    regions = survey['pixels']//subpixels
    survey['regions'] = regions

    # for each region compute the 1D P(k) and xi(r)
    Pk_region = np.zeros((regions,len(k_modes)),dtype=np.float32)
    xi_region = np.zeros((regions,len(r_modes)),dtype=np.float32)
    for i in range(regions):

        print('Computing P_1D(k) and xi_1D(r) of region',i,'out of',regions)

        # find the indexes of the region
        indexes_region = np.arange(i*subpixels,(i+1)*subpixels)
    
        ############################## Pk(k) ###############################
        # find the delta_k2 matrix of the region
        delta_k2_region = delta_k2[:,indexes_region]

        # take all LOS and compute the average value for each mode
        delta_k2_stacked = np.mean(delta_k2_region,dtype=np.float64,axis=1)

        # compute the 1D P(k)
        Pk = np.histogram(k,bins=k_bins,weights=delta_k2_stacked)[0]
        Pk = Pk/(BoxSize*k_modes);  Pk_region[i] = Pk;  del delta_k2_stacked

        ############################## xi(r) ###############################
        # Fourier transform P(k) to obtain xi(r)
        xi = scipy.fftpack.fftn(delta_k2_region,overwrite_x=True,axes=(0,))
        del delta_k2_region

        # take all LOS and compute the average value for each mode
        xi_stacked = np.mean(xi,dtype=np.float64,axis=1);  del xi

        # compute the 1D xi(r)
        xi = np.histogram(r,bins=r_bins,weights=xi_stacked)[0]
        xi = xi/r_modes;  xi_region[i] = xi;  del xi_stacked
    del indexes_region    

    # do a loop removing one region and compute P(k) and xi(r)
    Pk_jack = np.zeros((regions,len(k_modes)),dtype=np.float64)
    xi_jack = np.zeros((regions,len(r_modes)),dtype=np.float64)
    indexes = np.arange(regions)
    for i in range(regions):

        # find the indexes of all regions other than the deleted one
        indexes_to_consider = np.delete(indexes,i)

        # compute the average P(k) and xi(r) after removing the region
        Pk_jack[i] = np.mean(Pk_region[indexes_to_consider],axis=0)
        xi_jack[i] = np.mean(xi_region[indexes_to_consider],axis=0)

    # from all jackknife P(k) and xi(r) compute mean and variance
    Pk_mean = np.sum(Pk_jack,axis=0,dtype=np.float64)/regions
    xi_mean = np.sum(xi_jack,axis=0,dtype=np.float64)/regions

    Pk_var  = np.sqrt(np.sum((Pk_jack-Pk_mean)**2,axis=0,dtype=np.float64)*\
                          (regions-1.0)/regions)
    xi_var  = np.sqrt(np.sum((xi_jack-xi_mean)**2,axis=0,dtype=np.float64)*\
                          (regions-1.0)/regions)
    del Pk_jack, xi_jack, indexes

    return [k_bin,r_bin,Pk_mean,xi_mean,Pk_var,xi_var]
###############################################################################
###############################################################################
# This routine reads the cosmo,cosmo+noise and cosmo+noise+fg maps and computes 
# the 1D P(k) taking into account the mask
# root_map --------> array containing the roots of the 5 different maps
# fmask -----------> array containing the files with the different masks
def clustering_1D_all(root_map,nuTable,D,Omega_m,Omega_L,h,n_side_map,
                      bin_min,bin_max,T_instrument,time_total,fsky,n_dish,
                      fmask,f_out):

    # define cosmology dictionary
    cosmo = {'Omega_m':Omega_m, 'Omega_L':Omega_L, 'h':h}

    # define instrument dictionary
    instrument = {'D':D, 'n_dish':n_dish, 'T_instrument':T_instrument}

    # obtain some parameters of the survey
    survey = get_survey_params(bin_min,bin_max,nuTable,cosmo,instrument)
    survey['time_total'] = time_total
    survey['fsky']       = fsky
    dims = survey['dims'];  BoxSize = survey['BoxSize']
    
    # define the data structure: create five structures for maps: 1 for maps
    # with cosmo, 2 for maps with cosmo+noise and 2 for maps with cosmo+noise+fg
    pixels = HP.nside2npix(n_side_map)  #number of pixels in a map with n_side
    delta1 = np.zeros((dims,pixels),dtype=np.float32) #cosmo
    delta2 = np.zeros((dims,pixels),dtype=np.float32) #cosmo+noise1
    delta3 = np.zeros((dims,pixels),dtype=np.float32) #cosmo+noise2
    delta4 = np.zeros((dims,pixels),dtype=np.float32) #cosmo+fg+noise1
    delta5 = np.zeros((dims,pixels),dtype=np.float32) #cosmo+fg+noise2
    print('Number of pixels in the maps =',pixels,'\n')
    survey['pixels'] = pixels

    # compute T_system and sigma_T for each map and the time per pixel
    sigmaT(survey,instrument);  sigma_T = survey['sigma_T']  #mK
                                       
    # compute expected noise level and variance from system noise
    Tsystem_z = np.interp(survey['mean_z'],survey['z_map'][::-1],
                          survey['T_system'][::-1]) #mK
    noise_level_variance(cosmo,survey,instrument,survey['mean_z'],Tsystem_z)


    # read the maps and smooth them
    for i in range(bin_min, bin_max+1):
        
        # read maps
        suffix = str(i).zfill(3) + '.fits'
        map1 = HP.read_map(root_map[0]+suffix, nest=False, verbose=False)
        map2 = HP.read_map(root_map[1]+suffix, nest=False, verbose=False)
        map3 = HP.read_map(root_map[2]+suffix, nest=False, verbose=False)
        map4 = HP.read_map(root_map[3]+suffix, nest=False, verbose=False)
        map5 = HP.read_map(root_map[4]+suffix, nest=False, verbose=False)
        
        # compute map mean temperature, find delta_T and fill data matrix
        meanT = np.mean(map1,dtype=np.float64);  delta1[i-bin_min] = map1-meanT
        meanT = np.mean(map2,dtype=np.float64);  delta2[i-bin_min] = map2-meanT
        meanT = np.mean(map3,dtype=np.float64);  delta3[i-bin_min] = map3-meanT
        meanT = np.mean(map4,dtype=np.float64);  delta4[i-bin_min] = map4-meanT
        meanT = np.mean(map5,dtype=np.float64);  delta5[i-bin_min] = map5-meanT
        del map1,map2,map3,map4,map5


    # Fourier transform only along the given axis
    print('\nFFT the overdensity fields along axis 0')
    delta1_k = scipy.fftpack.fftn(delta1,overwrite_x=True,axes=(0,))
    delta1_k *= (BoxSize/dims);  del delta1

    delta2_k = scipy.fftpack.fftn(delta2,overwrite_x=True,axes=(0,))
    delta2_k *= (BoxSize/dims);  del delta2

    delta3_k = scipy.fftpack.ifftn(delta3,overwrite_x=True,axes=(0,))
    delta3_k *= BoxSize;         del delta3

    delta4_k = scipy.fftpack.fftn(delta4,overwrite_x=True,axes=(0,))
    delta4_k *= (BoxSize/dims);  del delta4

    delta5_k = scipy.fftpack.ifftn(delta5,overwrite_x=True,axes=(0,))
    delta5_k *= BoxSize;         del delta5

    # compute delta^2(k) for auto-correlation w/o noise and for 
    # cross-correlation with noise
    delta2_k_auto   = np.absolute(delta1_k)**2;    del delta1_k
    delta2_k_cross1 = np.real(delta2_k*delta3_k);  del delta2_k,delta3_k
    delta2_k_cross2 = np.real(delta4_k*delta5_k);  del delta4_k,delta5_k
    
    # define suffix of output files
    suffix = '%d-%d_z=%.2f_D=%.0f_t=%.0f_nside=%d.dat'\
        %(bin_min,bin_max,survey['mean_z'],D,time_total,n_side_map)

    ######################## cosmo ##########################
    # compute mean P(k) and xi(r) and determine variance with JK
    [k_bin,r_bin,Pk_mean,xi_mean,Pk_var,xi_var] = \
        clustering_region(cosmo,survey,instrument,delta2_k_auto,
                          fmask[0],do_JK=False);  del delta2_k_auto
                          
    # save results to file ignoring the DC mode
    f1 = f_out+'/Pk_cosmo_'+suffix;  f2 = f_out+'/xi_cosmo_'+suffix
    np.savetxt(f1,np.transpose([k_bin[1:],Pk_mean[1:],Pk_var[1:]]))
    np.savetxt(f2,np.transpose([r_bin[1:],xi_mean[1:],xi_var[1:]]))

    #################### cosmo+noise #######################
    # compute mean P(k) and xi(r) and determine variance with JK
    [k_bin,r_bin,Pk_mean,xi_mean,Pk_var,xi_var] = \
        clustering_region(cosmo,survey,instrument,delta2_k_cross1,
                          fmask[0],do_JK=False);  del delta2_k_cross1
                          
    # save results to file ignoring the DC mode
    f1 = f_out+'/Pk_noise_'+suffix;  f2 = f_out+'/xi_noise_'+suffix
    np.savetxt(f1,np.transpose([k_bin[1:],Pk_mean[1:],Pk_var[1:]]))
    np.savetxt(f2,np.transpose([r_bin[1:],xi_mean[1:],xi_var[1:]]))

    ################## cosmo+noise+fg ######################
    # compute mean P(k) and xi(r) and determine variance with JK
    [k_bin,r_bin,Pk_mean,xi_mean,Pk_var,xi_var] = \
        clustering_region(cosmo,survey,instrument,delta2_k_cross2,
                          fmask[1],do_JK=False);  del delta2_k_cross2
                          
    # save results to file ignoring the DC mode
    f1 = f_out+'/Pk_fg_'+suffix;  f2 = f_out+'/xi_fg_'+suffix
    np.savetxt(f1,np.transpose([k_bin[1:],Pk_mean[1:],Pk_var[1:]]))
    np.savetxt(f2,np.transpose([r_bin[1:],xi_mean[1:],xi_var[1:]]))
###############################################################################
###############################################################################


###############################################################################
###############################################################################
# This routine reads the healpix maps by David Alonso and computes the
# radial 21cm P(k) and xi(r).
# root_map -----------------------> 
# nuTable ------------------------> file with frequencies and redshifts of maps
# D ------------------------------> diameter of the antenna in m
# n_dish -------------------------> number of antennae
# T_instrument -------------------> temperature of the antenna in K
# time_total ---------------------> total observing time in hours
# fsky ---------------------------> fraction of the sky observed 
# Omega_m ------------------------> value of Omega_m
# Omega_L ------------------------> value of Omega_l
# h ------------------------------> value of h
# n_side_map ---------------------> 
# subpixels ----------------------> 
# bin_min ------------------------> 
# bin_max ------------------------> 
# f_out --------------------------> 
def clustering_1D(root_map,nuTable,D,Omega_m,Omega_L,h,n_side_map,subpixels,
                  bin_min,bin_max,T_instrument,time_total,fsky,n_dish,
                  fmask,do_JK,f_out):

    # define cosmology dictionary
    cosmo = {'Omega_m':Omega_m, 'Omega_L':Omega_L, 'h':h}

    # define instrument dictionary
    instrument = {'D':D, 'n_dish':n_dish, 'T_instrument':T_instrument}

    # obtain some parameters of the survey
    survey = get_survey_params(bin_min,bin_max,nuTable,cosmo,instrument)
    survey['time_total'] = time_total
    survey['fsky']       = fsky
    survey['subpixels']  = subpixels 
    dims = survey['dims'];  BoxSize = survey['BoxSize']

    # compute the growth factor at the mean redshift
    Dz = CL.linear_growth_factor(survey['mean_z'],Omega_m,Omega_L);  
    cosmo['Dz'] = Dz;  print('D(<z>=%.3f) = %.3f'%(survey['mean_z'],Dz));  del Dz
    
    # define the data structure: create three structures for maps: one for maps
    # with no noise, and two for maps with noise for cross-correlations
    pixels = HP.nside2npix(n_side_map)  #number of pixels in a map with n_side
    delta1 = np.zeros((dims,pixels),dtype=np.float32) #no noise
    delta2 = np.zeros((dims,pixels),dtype=np.float32) #with noise
    delta3 = np.zeros((dims,pixels),dtype=np.float32) #with noise
    print('Number of pixels in the maps =',pixels,'\n')
    survey['pixels'] = pixels

    # compute T_system and sigma_T for each map and the time per pixel
    sigmaT(survey,instrument);  sigma_T = survey['sigma_T']  #mK
                                       
    # compute expected noise level and variance from system noise
    Tsystem_z = np.interp(survey['mean_z'],survey['z_map'][::-1],
                          survey['T_system'][::-1]) #mK
    noise_level_variance(cosmo,survey,instrument,survey['mean_z'],Tsystem_z)


    # read the maps and smooth them
    for i in range(bin_min, bin_max+1):

        # get map name and read it
        print('working with map:',i)
        f_map = root_map    + str(i).zfill(3) + '.fits';  
        #map   = HP.read_map(f_map,nest=True,verbose=False)
        map   = HP.read_map(f_map,nest=False,verbose=False)
        map   = HP.smoothing(map,fwhm=survey['beam_fwhm'][i],regression=False)
        map   = HP.reorder(map,r2n=True)  #change ordering from ring to nest
        map   = HP.ud_grade(map,nside_out=n_side_map,
                            order_in='NESTED',order_out='NESTED')

        # plot to make sure that submaps are coherent
        #for i in range(pixels/subpixels):
        #    map2 = np.copy(map);  map2[i*subpixels:(i+1)*subpixels] = 0 
        #    HP.mollview(map2,nest=True)
        #show()

        # make three copies of the map: 1 without noise and 2 with noise
        # maps begin from 1, while sigma_T array begin from 0
        map1 = np.copy(map)                                                #mK
        map2 = np.copy(map);  map2 += sigma_T[i-1]*np.random.randn(pixels) #mK
        map3 = np.copy(map);  map3 += sigma_T[i-1]*np.random.randn(pixels) #mK

        # compute map mean temperature, find delta_T and fill data matrix
        mean_Tb1 = np.mean(map1,dtype=np.float64) #mK
        mean_Tb2 = np.mean(map2,dtype=np.float64) #mK
        mean_Tb3 = np.mean(map3,dtype=np.float64) #mK
        map1 = map1 - mean_Tb1;  delta1[i-bin_min] = map1
        map2 = map2 - mean_Tb2;  delta2[i-bin_min] = map2 
        map3 = map3 - mean_Tb3;  delta3[i-bin_min] = map3
        print('z = %.3f ----> <Tb> = %.4f mK'%(survey['z_map'][i],mean_Tb1))
        print('%.3f < delta_T(%d) < %.3f\n'%(np.min(map1),i,np.max(map1)))
        del map1,map2,map3,map


    # Fourier transform only along the given axis
    print('\nFFT the overdensity field along axis 0')
    delta1_k = scipy.fftpack.fftn(delta1,overwrite_x=True,axes=(0,))
    delta1_k *= (BoxSize/dims);  del delta1

    delta2_k = scipy.fftpack.fftn(delta2,overwrite_x=True,axes=(0,))
    delta2_k *= (BoxSize/dims);  del delta2

    delta3_k = scipy.fftpack.ifftn(delta3,overwrite_x=True,axes=(0,))
    delta3_k *= BoxSize;         del delta3

    # compute delta^2(k) for auto-correlation w/o noise and for 
    # cross-correlation with noise
    delta2_k_auto  = np.absolute(delta1_k)**2;    del delta1_k
    delta2_k_cross = np.real(delta2_k*delta3_k);  del delta2_k,delta3_k
    
    # define suffix of output files
    suffix = '%d-%d_z=%.2f_D=%.0f_t=%.0f_nside=%d.dat'\
        %(bin_min,bin_max,survey['mean_z'],D,time_total,n_side_map)

    #################### auto-correlation ########################
    # compute mean P(k) and xi(r) and determine variance with JK
    [k_bin,r_bin,Pk_mean,xi_mean,Pk_var,xi_var] = \
        clustering_region(cosmo,survey,instrument,delta2_k_auto,fmask,do_JK)
    del delta2_k_auto
                          
    # save results to file ignoring the DC mode
    f1 = f_out+'/Pk_new_cosmo_'+suffix;  f2 = f_out+'/xi_new_cosmo_'+suffix
    np.savetxt(f1,np.transpose([k_bin[1:],Pk_mean[1:],Pk_var[1:]]))
    np.savetxt(f2,np.transpose([r_bin[1:],xi_mean[1:],xi_var[1:]]))

    #################### cross-correlation #######################
    # compute mean P(k) and xi(r) and determine variance with JK
    [k_bin,r_bin,Pk_mean,xi_mean,Pk_var,xi_var] = \
        clustering_region(cosmo,survey,instrument,delta2_k_cross,fmask,do_JK)
    del delta2_k_cross
                          
    # save results to file ignoring the DC mode
    f1 = f_out+'/Pk_new_noise_'+suffix;  f2 = f_out+'/xi_new_noise_'+suffix
    np.savetxt(f1,np.transpose([k_bin[1:],Pk_mean[1:],Pk_var[1:]]))
    np.savetxt(f2,np.transpose([r_bin[1:],xi_mean[1:],xi_var[1:]]))
    
    print('Noise level = %.3f Mpc/h'%survey['noise_level_cross'])
    print('sigma noise = %.3e Mpc/h'%survey['sigma_noise_cross'])
###############################################################################
###############################################################################


###############################################################################
###############################################################################
def clustering_1D_fg(root_map1,root_map2,nuTable,D,Omega_m,Omega_L,h,n_side_map,
                     subpixels,bin_min,bin_max,T_instrument,time_total,fsky,
                     n_dish,method,n_fog,fmask,do_JK,f_out):
                  
    # define cosmology dictionary
    cosmo = {'Omega_m':Omega_m, 'Omega_L':Omega_L, 'h':h}

    # define instrument dictionary
    instrument = {'D':D, 'n_dish':n_dish, 'T_instrument':T_instrument}

    # obtain some parameters of the survey
    survey = get_survey_params(bin_min,bin_max,nuTable,cosmo,instrument)
    survey['time_total'] = time_total
    survey['fsky']       = fsky
    survey['subpixels']  = subpixels 
    dims = survey['dims'];  BoxSize = survey['BoxSize']

    # compute the growth factor at the mean redshift
    Dz = CL.linear_growth_factor(survey['mean_z'],Omega_m,Omega_L);  
    cosmo['Dz'] = Dz;  print('D(<z>=%.3f) = %.3f'%(survey['mean_z'],Dz));  del Dz
    
    # define the data structure: create three structures for maps: one for maps
    # with no noise, and two for maps with noise for cross-correlations
    pixels = HP.nside2npix(n_side_map)  #number of pixels in a map with n_side
    delta1 = np.zeros((dims,pixels),dtype=np.float32) #map1
    delta2 = np.zeros((dims,pixels),dtype=np.float32) #map2
    print('Number of pixels in the maps =',pixels,'\n')
    survey['pixels'] = pixels

    # compute T_system and sigma_T for each map and the time per pixel
    sigmaT(survey,instrument)
    sigma_T = survey['sigma_T']  #mK
                                       
    # compute expected noise level and variance from system noise
    Tsystem_z = np.interp(survey['mean_z'],survey['z_map'][::-1],
                          survey['T_system'][::-1]) #mK
    noise_level_variance(cosmo,survey,instrument,survey['mean_z'],Tsystem_z)


    # read the maps and smooth them
    for i in range(bin_min, bin_max+1):

        # get map name and read it
        print('working with map:',i)
        f_map1 = root_map1 + str(i).zfill(3) + '.fits'
        f_map2 = root_map2 + str(i).zfill(3) + '.fits'  
        map1   = HP.read_map(f_map1,nest=False,verbose=False)
        map2   = HP.read_map(f_map2,nest=False,verbose=False)
        
        # compute map mean temperature, find delta_T and fill data matrix
        # when cleaning the foregrounds the map only contain T-<T>
        mean_T1 = np.mean(map1,dtype=np.float64)
        mean_T2 = np.mean(map2,dtype=np.float64)
        delta1[i-bin_min] = map1-mean_T1
        delta2[i-bin_min] = map2-mean_T2
        del map1,map2


    # Fourier transform only along the given axis
    print('\nFFT the overdensity field along axis 0')
    delta1_k = scipy.fftpack.fftn(delta1,overwrite_x=True,axes=(0,))
    delta1_k *= (BoxSize/dims);  del delta1

    delta2_k = scipy.fftpack.ifftn(delta2,overwrite_x=True,axes=(0,))
    delta2_k *= BoxSize;         del delta2

    # compute delta^2(k) for cross-correlation with noise
    delta2_k_cross = np.real(delta1_k*delta2_k);  del delta1_k,delta2_k
    
    #################### cross-correlation #######################
    # compute mean P(k) and xi(r) and determine variance with JK
    data = clustering_region(cosmo,survey,instrument,delta2_k_cross,fmask,do_JK)
    if do_JK:  [k_bin,r_bin,Pk_mean,xi_mean,Pk_var,xi_var] = data
    else:      [k_bin,r_bin,Pk_mean,xi_mean] = data
    del delta2_k_cross
                          
    # save results to file ignoring the DC mode
    suffix = '%d-%d_z=%.2f_D=%.0f_t=%.0f_nside=%d.dat'\
        %(bin_min,bin_max,survey['mean_z'],D,time_total,n_side_map)
    f1 = f_out+'/Pk_fg_'+method+'_'+n_fog+'_'+suffix
    f2 = f_out+'/xi_fg_'+method+'_'+n_fog+'_'+suffix
    if do_JK:
        np.savetxt(f1,np.transpose([k_bin[1:],Pk_mean[1:],Pk_var[1:]]))
        np.savetxt(f2,np.transpose([r_bin[1:],xi_mean[1:],xi_var[1:]]))
    else:
        np.savetxt(f1,np.transpose([k_bin[1:],Pk_mean[1:]]))
        np.savetxt(f2,np.transpose([r_bin[1:],xi_mean[1:]]))
    
    print('Noise level = %.3f Mpc/h'%survey['noise_level_cross'])
    print('sigma noise = %.3e Mpc/h'%survey['sigma_noise_cross'])
###############################################################################
###############################################################################


###############################################################################
###############################################################################
# This routine carry out a fit to the P(k) data in f_in, over the interval
# [kmin,kmax] using as template the P(k) in f_w and f_nw (which are the mean 
# over 100 realizations with wiggles and no-wiggles). The routine requires an
# initial guess of the parameters (theta0)
# f_w ---------> file with the P(k)_wiggles template
# f_nw --------> file with the P(k)_no_wiggles template
# f_in --------> file with P(k) to fit
# kmin,kmax ---> k-range used for the fit
# theta0 ------> initial guess of best fit parameters
# ndim --------> number of free parameters to fit
# nwalkers ----> number of MCMC chains
# chain_pts ---> number of points in each MCMC chain
# fout --------> name of file to save MCMC plots
def fitting(f_w,f_nw,f_in,kmin,kmax,theta0,ndim,nwalkers,chain_pts,fout):
    
    # sanity check
    if ndim!=len(theta0):
        print('number of free parameters is not %d!!!'%ndim);  sys.exit()

    # read templates and realization
    kw,Pkw,dPkw    = np.loadtxt(f_w,unpack=True)
    knw,Pknw,dPknw = np.loadtxt(f_nw,unpack=True)
    k,Pk,dPk       = np.loadtxt(f_in,unpack=True)

    # keep only with data in the given k-range
    indexes  = np.where((k>kmin) & (k<kmax))[0]
    k,Pk,dPk = k[indexes],Pk[indexes],dPkw[indexes] #k,Pk and dPk of data to fit

    # compute number of degrees of freedom
    ndof = len(indexes)-len(theta0);  del indexes

    ###### find best fit parameters and minimum chi2 ######
    chi2_func = lambda *args: -lnlike_sim_template(*args)
    best_fit  = minimize(chi2_func,theta0,args=(k,Pk,dPk,kw,Pkw,knw,Pknw),
                         method='Powell')
    theta_best_fit = best_fit["x"]
    chi2 = chi2_func(theta_best_fit,k,Pk,dPk,kw,Pkw,knw,Pknw)*1.0/ndof
    
    alpha, b, p0, p1, p2 = theta_best_fit
    print('alpha = %10.3e'%alpha)
    print('b     = %10.3e'%b)
    print('p0    = %10.3e'%p0)
    print('p1    = %10.3e'%p1)
    print('p2    = %10.3e'%p2)
    print('chi2  = %10.3f\n'%chi2)
    #######################################################

    ###################### run MCMC #######################
    # starting point of each MCMC chain: small ball around best fit or estimate
    pos = [theta_best_fit+0.2*np.random.randn(ndim) for i in range(nwalkers)]

    # run MCMC
    sampler = emcee.EnsembleSampler(nwalkers,ndim,lnlike_sim_template,
                                args=(k,Pk,dPk,kw,Pkw,knw,Pknw))      
    sampler.run_mcmc(pos,chain_pts)
    sampler.reset()

    # get the points in the MCMC chains
    samples = sampler.chain[:,500:,:].reshape((-1,ndim))
    sampler.reset()

    # make the plot and save it to file
    fig = corner.corner(samples,truths=theta0,
                        labels=[r"$\alpha$",r"$b$",r"$p_0$",r"$p_1$",r"$p_2$"])
    fig.savefig(fout)

    # compute best fit and associated error
    alpha, b, p0, p1, p2 =  map(lambda v: (v[1], v[2]-v[1], v[1]-v[0]),
                                zip(*np.percentile(samples, [16, 50, 84],
                                                   axis=0)))
                                      
    # compute chi^2 of best fit
    theta = [alpha[0],b[0],p0[0],p1[0],p2[0]]
    chi2 = chi2_func(theta,k,Pk,dPk,kw,Pkw,knw,Pknw)/ndof

    print('alpha = %10.3e + %.2e - %.2e'%(alpha[0],alpha[1],alpha[2]))
    print('b     = %10.3e + %.2e - %.2e'%(b[0],b[1],b[2]))
    print('p0    = %10.3e + %.2e - %.2e'%(p0[0],p0[1],p0[2]))
    print('p1    = %10.3e + %.2e - %.2e'%(p1[0],p1[1],p1[2]))
    print('p2    = %10.3e + %.2e - %.2e'%(p2[0],p2[1],p2[2]))
    print('chi2  = %10.3f'%chi2)

    # compute the P(k) of the model
    Pk = model(kw,theta,kw,Pkw,knw,Pknw)

    return alpha,b,p0,p1,p2,chi2,Pk
###############################################################################
###############################################################################



###############################################################################
###############################################################################
# This function makes a fit to the data, inputed as (k_real,Pk_real,dPk_real)
# using a theoretical template
def fitting_theory(theta0,k_real,Pk_real,dPk_real,kw,Pkw,Pknw,beta,Radii,
                   nwalkers,chain_pts,f_contours):

    # compute number of parameters and degrees of freedom
    ndim = len(theta0);  ndof = len(Pk_real) - ndim

    ###### find best fit parameters and minimum chi2 ######
    chi2_func = lambda *args: -lnlike_theory(*args)
    best_fit  = minimize(chi2_func,theta0,
                         args=(k_real,Pk_real,dPk_real,
                               kw,Pkw,Pknw,beta,Radii),method='Powell')
    theta_best_fit = best_fit["x"]
    chi2 = chi2_func(theta_best_fit,k_real,Pk_real,dPk_real,
                     kw,Pkw,Pknw,beta,Radii)*1.0/ndof
    
    alpha, b, R, p0, p1, p2 = theta_best_fit
    print('alpha = %10.3e'%alpha)
    print('b     = %10.3e'%b)
    print('R     = %10.3e'%R)
    print('p0    = %10.3e'%p0)
    print('p1    = %10.3e'%p1)
    print('p2    = %10.3e'%p2)
    #print('p3    = %10.3e'%p3)
    print('chi2  = %10.3f\n'%chi2)
    #######################################################                 

    ###################### run MCMC #######################
    # starting point of each MCMC chain: small ball around best fit or estimate
    pos = [theta_best_fit*(1.0+0.1*np.random.randn(ndim)) for i in range(nwalkers)]

    # run MCMC
    sampler = emcee.EnsembleSampler(nwalkers,ndim,lnlike_theory,
                                    args=(k_real,Pk_real,dPk_real,
                                          kw,Pkw,Pknw,beta,Radii))
    sampler.run_mcmc(pos,chain_pts);  del pos

    # get the points in the MCMC chains
    samples = sampler.chain[:,6000:,:].reshape((-1,ndim))
    sampler.reset()

    # make the plot and save it to file
    fig = corner.corner(samples,truths=theta0,labels=[r"$\alpha$",
                                                      r"$b$",r"$R$",
                                                      r"$p_0$",r"$p_1$",
                                                      r"$p_2$"])
    fig.savefig(f_contours)

    # compute best fit and associated error
    alpha, b, R, p0, p1, p2 =  map(lambda v: (v[1], v[2]-v[1], v[1]-v[0]),
                                   zip(*np.percentile(samples, [16, 50, 84],
                                                      axis=0)))
    del samples, sampler
                                      
    # compute chi^2 of best fit
    theta_mcmc = [alpha[0],b[0],R[0],p0[0],p1[0],p2[0]]
    chi2 = chi2_func(theta_mcmc,k_real,Pk_real,dPk_real,kw,Pkw,Pknw,
                     beta,Radii)*1.0/ndof
                     

    # show results of MCMC
    print('alpha = %10.3e + %.2e - %.2e'%(alpha[0],alpha[1],alpha[2]))
    print('b     = %10.3e + %.2e - %.2e'%(b[0],b[1],b[2]))
    print('R     = %10.3e + %.2e - %.2e'%(R[0],R[1],R[2]))
    print('p0    = %10.3e + %.2e - %.2e'%(p0[0],p0[1],p0[2]))
    print('p1    = %10.3e + %.2e - %.2e'%(p1[0],p1[1],p1[2]))
    print('p2    = %10.3e + %.2e - %.2e'%(p2[0],p2[1],p2[2]))
    #print('p3    = %10.3e + %.2e - %.2e'%(p3[0],p3[1],p3[2]))
    print('chi2  = %10.3f'%chi2)

    # compute best fit model for data and save results to file
    Pk_model = model_theory(k_real,theta_mcmc,kw,Pkw,Pknw,beta,Radii)
    np.savetxt('Best_fit.txt',np.transpose([k_real,Pk_model]))

    # compute best fit model without the polynomial and save results to file
    theta_mcmc = [alpha[0], b[0], R[0], 0.0, 0.0, 0.0]
    Pk_model = model_theory(k_real,theta_mcmc,kw,Pkw,Pknw,beta,Radii)
    np.savetxt('Best_fit_zero_pol.txt',np.transpose([k_real,Pk_model]))

    return alpha,b,R,p0,p1,p2,chi2
###############################################################################
###############################################################################


###############################################################################
###############################################################################
# This function computes P(k) in k values using as template the mean P(k)
# from the simulations
# P_model(k) = b*[(P_lin(k/a)-P_nw(k/a)) + P_nw(k)] + (p0*k + p1 + p2/k)
def model_sim_template(k,theta,kw,Pkw,knw,Pknw):
    alpha,b,p = theta[0],theta[1],theta[2:]
    Pnw  = np.interp(k,knw,Pknw)
    Pwa  = np.interp(k/alpha,kw,Pkw)
    Pnwa = np.interp(k/alpha,knw,Pknw)
    pol  = p[0]*k + p[1] + p[2]/k
    return b*((Pwa-Pnwa)+Pnw) + pol

# This functions returns the log-likelihood for the model with the sim template
def lnlike_sim_template(theta,k,Pk,dPk,kw,Pkw,knw,Pknw,beta):
    if (theta[0]<=0) or (theta[0]>5.0) or (theta[1]<=0):
        return -np.inf
    else:
        Pk_model = model(k,theta,kw,Pkw,knw,Pknw)
        return -np.sum(((Pk-Pk_model)/dPk)**2,dtype=np.float64)

# This function computes P(k) in k values using the theoretical template
# P_model(k) = b*[(P_w(k/a)-P_nw(k/a)) + P_nw(k)] + (p0*k + p1 + p2/k)
def model_theory(k,theta,kw,Pkw,Pknw,beta,Radii):

    alpha, bias, R, p = theta[0], theta[1], theta[2], theta[3:]        

    Rmin = Radii[0];  Rmax = Radii[-1];  R_bins = len(Radii)
    imin = int((R-Rmin)*(R_bins-1.0)/(Rmax-Rmin)) #index of closest array
    if imin==R_bins:      imin = R_bins-1
    if imin==(R_bins-1):  imax = imin
    else:                 imax = imin + 1

    if imin==imax:  
        print('error!!!!');  sys.exit()

    Rdown = (Rmax-Rmin)/(R_bins-1.0)*1.0*imin + Rmin
    Rup   = (Rmax-Rmin)/(R_bins-1.0)*1.0*imax + Rmin

    Pkw_model  = (Pkw[imax]-Pkw[imin])/(Rup-Rdown)*(R-Rdown) + Pkw[imin]
    Pknw_model = (Pknw[imax]-Pknw[imin])/(Rup-Rdown)*(R-Rdown) + Pknw[imin]

    Pnw  = np.interp(k,       kw, Pknw_model)
    Pwa  = np.interp(k/alpha, kw, Pkw_model)
    Pnwa = np.interp(k/alpha, kw, Pknw_model)

    return bias**2*((Pwa-Pnwa)+Pnw) + p[0]*k + p[1] + p[2]/k

# This functions returns the log-likelihood for the theory model
# theta -------> array with parameters to fit [alpha,bias,R,p0,p1,p2]
# k,Pk,dPk ----> data to fit
# kw ----------> array containing the values of k of the templates
# Pkw ---------> matrix with values of P_wiggles(k) for different R values
# Pknw --------> matrix with values of P_no_wiggles(k) for different R values
# beta --------> value of beta
def lnlike_theory(theta,k,Pk,dPk,kw,Pkw,Pknw,beta,Radii):
    # put priors here
    if (theta[0]<0.8) or (theta[0]>1.2) or (theta[1]<=0.0) \
            or (theta[2]>=Radii[-1]) or (theta[2]<=Radii[0]):
            #or (theta[6]<0.0):
        return -np.inf
    else:
        Pk_model = model_theory(k,theta,kw,Pkw,Pknw,beta,Radii)
        chi2 = -np.sum(((Pk-Pk_model)/dPk)**2,dtype=np.float64)
        return chi2

###############################################################################
###############################################################################


# This routine computes the mean and variance of the results
def mean_variance_Pk(f_3D_Pk,nuTable,nside,bin_min,bin_max,D,time_total,
                     realizations,Omega_m,Omega_L,h,root_fout,do_nw,do_fg):

    # compute Gaussian errors
    mean_z,k_g,dPk_g = Gaussian_1D_errors(f_3D_Pk,nuTable,bin_min,bin_max,
                                          Omega_m,Omega_L,h,D)

    # read the frequencies and redshifts of the maps
    num, nui, nuf, zi, zf = np.loadtxt(nuTable,unpack=True);  z = 0.5*(zi+zf)

    # compute the mean redshift of the survey
    mean_z = 0.5*(z[bin_min-1]+z[bin_max-1]);  print('survey <z> = %.3f'%mean_z)

    suffix = '%d-%d_z=%.2f_D=%.0f_t=%.0f_nside=%d.dat'\
        %(bin_min,bin_max,mean_z,D,time_total,nside)

    Pk_cosmo, Pk_noise, Pk_nw_cosmo, Pk_fg = [],[],[],[]
    for i in np.arange(1,1+realizations):
        
        f1 = root_fout+str(i)+'/Pk_cosmo_'+suffix
        f2 = root_fout+str(i)+'/Pk_noise_'+suffix
        if do_nw:  f3 = root_fout+'nw_'+str(i)+'/Pk_cosmo_'+suffix
        if do_fg:  f4 = root_fout+str(i)+'/Pk_fg_'+suffix

        k,Pk,dumb = np.loadtxt(f1,unpack=True);  Pk_cosmo.append(Pk)
        k,Pk,dumb = np.loadtxt(f2,unpack=True);  Pk_noise.append(Pk)
        if do_nw:
            k,Pk,dumb = np.loadtxt(f3,unpack=True);  Pk_nw_cosmo.append(Pk)
        if do_fg:  k,Pk,dumb = np.loadtxt(f4,unpack=True);  Pk_fg.append(Pk)
        
    Pk_cosmo = np.array(Pk_cosmo);        Pk_noise = np.array(Pk_noise)
    Pk_nw_cosmo = np.array(Pk_nw_cosmo);  Pk_fg = np.array(Pk_fg)

    f = open(root_fout+'mean_Pk_cosmo_'+suffix,'w')
    g = open(root_fout+'mean_Pk_noise_'+suffix,'w')
    if do_nw:  h = open(root_fout+'mean_Pk_nw_cosmo_'+suffix,'w')
    if do_fg:  l = open(root_fout+'mean_Pk_fg_'+suffix,'w')
    for i in range(len(k)):
        f.write(str(k[i])+' '+str(np.mean(Pk_cosmo[:,i]))+' '+\
                    str(np.std(Pk_cosmo[:,i]))+' '+str(dPk_g[i])+'\n')
        g.write(str(k[i])+' '+str(np.mean(Pk_noise[:,i]))+' '+\
                    str(np.std(Pk_noise[:,i]))+'\n')
        if do_nw:
            h.write(str(k[i])+' '+str(np.mean(Pk_nw_cosmo[:,i]))+' '+\
                        str(np.std(Pk_nw_cosmo[:,i]))+'\n')
        if do_fg:
            l.write(str(k[i])+' '+str(np.mean(Pk_fg[:,i]))+' '+\
                        str(np.std(Pk_fg[:,i]))+'\n')
    f.close();  g.close()
    if do_nw: h.close()
    if do_fg: l.close()    

###############################################################################
###############################################################################

# compute P_1D(k) = \int_0^oo k_perp P_3D(k_par,k_perp) e^(-k_perp^2R^2) dk_perp
def Pk_1D_integrand(y,x,k,Pk,k_par,beta,R):
    k_mod     = math.sqrt(k_par**2 + x**2)
    Pk_interp = np.interp(np.log(k_mod),np.log(k),Pk)
    aux       = math.exp(-(x*R)**2 - 4.0*k_mod**2)
    #aux       = math.exp(-(x*R)**2)
    return Pk_interp*x*(1.0+beta*(k_par/k_mod)**2)**2*aux

# This function computes the 1D P(k) within [kmin,kmax] from the 3D one 
# taking into account the value of beta and R
def Pk_1D_model(kmin,kmax,bins_1D,k,Pk,beta,R):
    k_1D  = np.logspace(np.log10(kmin),np.log10(kmax),bins_1D)
    Pk_1D = np.zeros(bins_1D,dtype=np.float64)
    for i,k_par in enumerate(k_1D):
        yinit = [0.0];  k_limits = [0.0,np.sqrt(k[-1]**2-k_par**2)]
        Pk_1D[i] = si.odeint(Pk_1D_integrand,yinit,k_limits,
                             args=(k,Pk,k_par,beta,R),
                             mxstep=100000,rtol=1e-10,atol=1e-12,
                             h0=1e-10)[1][0]/(2*np.pi)
    return [k_1D,Pk_1D]


# This is the main function used to fit the results of the simulations to the
# theoretical template
def fit_function(nuTable,bin_min,bin_max,model,kmin,kmax,k_bins,R_bins,
                 D,time_total,nside,Omega_m,Omega_L,h,f_3D_Pkw,f_3D_Pknw,
                 nwalkers,chain_pts,realizations):
                 

    # read frequencies and redshifts of the maps and find z_min,z_max and <z>
    num, nui, nuf, zi, zf = np.loadtxt(nuTable,unpack=True);  z = 0.5*(zi+zf)
    z_min = z[bin_max-1];  z_max = z[bin_min-1];  mean_z = 0.5*(z_min + z_max)

    # file to save results
    suffix = '%d-%d_z=%.2f_D=%.0f_t=%.0f_nside=%d.dat'\
        %(bin_min,bin_max,mean_z,D,time_total,nside)
    fout_partial = 'MCMC_'+model+'_'+suffix+'_'+str(myrank)

    # compute the value of the HI bias and 21cm bias
    HI_bias = HI_bias_model(mean_z)
    b_21cm  = bias_21cm_model(mean_z,Omega_m,Omega_L,h)

    # compute redshift-space distortion parameter beta
    Omega_mz = Omega_m*(1.0+mean_z)**3/(Omega_m*(1.0+mean_z)**3+Omega_L)
    beta     = (Omega_mz**0.54545454)/HI_bias

    # compute maximum and minimum comoving angular smoothings: use fwhm factor
    fwhm = 2.0*np.sqrt(2.0*np.log(2.0))  #fwhm = sigma*2*sqrt(2*log(2))
    Rmin = (0.21*(1.0+z_min)/D/fwhm)*CL.comoving_distance(z_min,Omega_m,Omega_L)
    Rmax = (0.21*(1.0+z_max)/D/fwhm)*CL.comoving_distance(z_max,Omega_m,Omega_L)

    # verbose
    if myrank==0:
        print('Expected bias = %.3f'%b_21cm)
        print('angular smoothing within [%.2f,%.2f] Mpc/h'%(Rmin,Rmax))
        print('mean angular smoothing = %.2f Mpc/h'%(0.5*(Rmin+Rmax)))

    # compute/read Pkw and Pknw matrices
    f_data = 'data_kmin=%.3f_kmax=%.3f_kbins=%d_Rbins=%d_'\
        %(kmin,kmax,k_bins,R_bins)+suffix
    if os.path.exists(f_data+'.npy'):
        kw,Radii,Pkw,Pknw = np.load(f_data+'.npy')
    else:
        # read the CAMB 3D P(k) with and without wiggles
        k3Dw, Pk3Dw  = np.loadtxt(f_3D_Pkw, unpack=True)
        k3Dnw,Pk3Dnw = np.loadtxt(f_3D_Pknw,unpack=True)

        Pkw    = np.zeros((R_bins,k_bins),dtype=np.float64)
        Pknw   = np.zeros((R_bins,k_bins),dtype=np.float64)
        Pkw_d  = np.zeros((R_bins,k_bins),dtype=np.float64)
        Pknw_d = np.zeros((R_bins,k_bins),dtype=np.float64)
        Radii  = np.linspace(Rmin,Rmax,R_bins)
        bins   = np.where((np.arange(R_bins)%nprocs)==myrank)[0]
        for i in bins:
            print(i)
            kw,Pkw_d[i]  = \
                Pk_1D_model(kmin,kmax,k_bins,k3Dw,Pk3Dw,beta,Radii[i])
            kw,Pknw_d[i] = \
                Pk_1D_model(kmin,kmax,k_bins,k3Dnw,Pk3Dnw,beta,Radii[i])
        comm.Barrier()
        for i in range(R_bins):
            comm.Reduce(Pkw_d[i],  Pkw[i],  op=MPI.SUM, root=0)   
            comm.Reduce(Pknw_d[i], Pknw[i], op=MPI.SUM, root=0)
        del Pkw_d,Pknw_d,k3Dw,Pk3Dw,k3Dnw,Pk3Dnw;  
        if myrank==0:  np.save(f_data,[kw,Radii,Pkw,Pknw])
        Pkw = comm.Bcast(Pkw,root=0);  Pknw = comm.Bcast(Pknw,root=0)
        
        
    # these are the realizations each cpu work on
    numbers = np.where((np.arange(realizations)%nprocs)==myrank)[0]+1

    # do a loop over all realizations
    f = open(fout_partial,'w')
    for i in numbers:

        print('Working with realization',i)

        # Power spectrum of the realization and errors
        folder           = 'results/'+str(i)
        f_Pk_errors      = 'results/mean_Pk_'+model+'_'+suffix
        f_Pk_realization = folder+'/Pk_'+model+'_'+suffix
        f_contours       = 'results/contours_'+model+' '+suffix[:-3]+'png'

        # read 1D P(k) of the realization and error
        k_real,Pk_real,dumb = np.loadtxt(f_Pk_realization,unpack=True)
        data                = np.loadtxt(f_Pk_errors,unpack=False)
        dPk_real            = data[:,2]

        # keep only data in k-range: for the fit we use k_real,Pk_real,dPk_real
        indexes  = np.where((k_real>kmin) & (k_real<kmax))[0]
        k_real   = k_real[indexes];  Pk_real = Pk_real[indexes]
        dPk_real = dPk_real[indexes];  del indexes

        # initial guess for the fit:  theta0 = [alpha, bias, R, p0, p1, p2]
        theta0 = [1.0, b_21cm, 0.5*(Rmin+Rmax), 0.0, 0.0, 0.0]

        # do the fitting and compute error on parameters
        data = fitting_theory(theta0,k_real,Pk_real,dPk_real,kw,Pkw,Pknw,
                              beta,Radii,nwalkers,chain_pts,f_contours)
        gc.collect()
                                  
        alpha,b,R,p0,p1,p2,chi2 = data
        del k_real,Pk_real,dPk_real,dumb

        # save results to file
        f.write(str(i)+' '+\
                str(alpha[0])+' '+str(alpha[1])+' '+str(alpha[2])+' '+\
                str(b[0])+' '+str(b[1])+' '+str(b[2])+' '+\
                str(R[0])+' '+str(R[1])+' '+str(R[2])+' '+\
                str(p0[0])+' '+str(p0[1])+' '+str(p0[2])+' '+\
                str(p1[0])+' '+str(p1[1])+' '+str(p1[2])+' '+\
                str(p2[0])+' '+str(p2[1])+' '+str(p2[2])+' '+\
                #str(p3[0])+' '+str(p3[1])+' '+str(p3[2])+' '+\
                str(chi2)+'\n')  
    f.close()

    # synchronize all processes here
    comm.Barrier()

    # merge the results from partial files
    if myrank==0:

        # concatenate all files and remove them
        fout = 'MCMC_'+model+'_'+suffix;  string = ''
        for i in range(nprocs):
            string += (fout+'_'+str(i)+' ')
        os.system('cat '+string+' > '+fout);  os.system('rm '+string)
    
        # read data from concatenated file, sort it and rewrite it
        f = open(fout,'r'); data = []
        for line in f.readlines():  
            data.append(line.split())
        f.close();  data = sorted(data,key=lambda x:int(x[0]))
        np.savetxt(fout,data,fmt='%s')

    # synchronize all processes here
    comm.Barrier()
###############################################################################
###############################################################################

# I = \int_0^oo dk_perp k_perp P^2(k_par,k_perp)
def Pk_1D_error(y,x,k,Pk,k_par,b_21,beta,R_smooth):
    k_mod = np.sqrt(x**2 + k_par**2)
    Pk_kpar_kperp = b_21**2*(1.0+beta*(k_par/k_mod)**2)**2*np.interp(k_mod,k,Pk)
    W = np.exp(-x**2*R_smooth**2/2.0)
    return x*(W**2*Pk_kpar_kperp)**2

# This routine computes the Gaussian errors on the 1D power spectrum 
def Gaussian_1D_errors(f_3D_Pk,nuTable,bin_min,bin_max,Omega_m,Omega_L,h,D):

    # read the 3D P(k) 
    k_3D,Pk_3D = np.loadtxt(f_3D_Pk,unpack=True)

    # read the frequencies and redshifts of the maps
    num, nui, nuf, zi, zf = np.loadtxt(nuTable,unpack=True);  z = 0.5*(zi+zf)

    # compute the mean redshift of the survey
    mean_z = 0.5*(z[bin_min-1]+z[bin_max-1]);  print('survey <z> = %.2f'%mean_z)

    # compute the comoving distance to the survey mean redshift
    d_co = CL.comoving_distance(mean_z,Omega_m,Omega_L)  #Mpc/h
    print('comoving distance to <z> = %.2f -------> %.2f Mpc/h'%(mean_z,d_co))
    
    # compute the HI bias and Tb at <z>
    Omega_HI = Omega_HI_model(mean_z)
    b_HI     = HI_bias_model(mean_z)                     #HI bias
    b_21     = bias_21cm_model(mean_z,Omega_m,Omega_L,h) #21cm bias

    # compute instrument angular resolution at <z> and smoothing radius 
    theta    = 0.21*(1.0+mean_z)/D/2.355
    R_smooth = d_co*theta
    print('Smoothing angular scale at <z> = %.2f = %.2f Mpc/h'%(mean_z,R_smooth))

    # compute the redshift-space distortion parameters beta
    Omega_mz = Omega_m*(1.0+mean_z)**3/(Omega_m*(1.0+mean_z)**3+Omega_L)
    beta     = Omega_mz**0.54545454/b_HI

    # compute maximum and minimum comoving distance and survey volune
    d_max = CL.comoving_distance(z[bin_min-1],Omega_m,Omega_L) #Mpc/h
    d_min = CL.comoving_distance(z[bin_max-1],Omega_m,Omega_L) #Mpc/h
    V     = 4.0/3.0*np.pi*(d_max**3-d_min**3)                  #(Mpc/h)^3

    # define the k-bins
    dims    = bin_max - bin_min + 1
    BoxSize = d_max-d_min  #Mpc/h
    k_bins  = np.linspace(0,dims//2,dims//2+1)*(2.0*np.pi/BoxSize)

    # compute the values of k and r for the 1D P(k) and xi(r)
    modes   = np.arange(dims,dtype=np.float64);  middle = dims//2
    indexes = np.where(modes>middle)[0];  modes[indexes] = modes[indexes]-dims
    k = modes*(2.0*np.pi/BoxSize)  #k in h/Mpc
    k = np.absolute(k);  #just take the modulus
    del indexes, modes

    # compute the number of modes and the average number-weighted value of k,r
    k_modes = np.histogram(k,bins=k_bins)[0]
    k_bin   = np.histogram(k,bins=k_bins,weights=k)[0]/k_modes

    # for each k_par make David's integral
    dPk = np.zeros(len(k_bin),dtype=np.float32)
    for i,k_par in enumerate(k_bin):
        k_perp_max = np.sqrt(k_3D[-1]**2-k_par**2)
        yinit = [0.0];  k_limits = [0, k_perp_max]
        delta_k = k_bins[i+1]-k_bins[i]
        dPk[i] = si.odeint(Pk_1D_error,yinit,k_limits,
                           args=(k_3D,Pk_3D,k_par,b_21,beta,R_smooth),
                           mxstep=10000,rtol=1e-12,atol=1e-15,
                           h0=1e-15)[1][0]/(delta_k*V)
        dPk[i] = np.sqrt(dPk[i])

    # avoid fundamental frequency
    k_bin = k_bin[1:];  dPk = dPk[1:]
    
    return [mean_z,k_bin,dPk]
###############################################################################
###############################################################################



import numpy as np 
import time,sys,os
cimport numpy as np
import scipy.integrate as SI
cimport cython
from libc.math cimport sqrt,pow,sin,cos,floor,fabs,round
cimport MAS_c as MASC

ctypedef MASC.FLOAT FLOAT

def FLOAT_type():
    if sizeof(MASC.FLOAT)==4:  return np.float32
    else:                      return np.float64


################################################################################
################################# ROUTINES #####################################
#### from particle positions, with optional weights, to 3D density fields ####
# MA(pos,number,BoxSize,MAS='CIC',W=None) ---> main routine
# NGP(pos,number,BoxSize)
# CIC(pos,number,BoxSize)
# TSC(pos,number,BoxSize)
# PCS(pos,number,BoxSize)
# NGPW(pos,number,BoxSize,W)
# CICW(pos,number,BoxSize,W)
# TSCW(pos,number,BoxSize,W)
# PCSW(pos,number,BoxSize,W)
# SPH_NGP(density,pos,radius,r_bins,part_in_shell,BoxSize,verbose)
# SPH_NGPW(density,pos,radius,W,r_bins,part_in_shell,BoxSize,verbose)
# NGPW_d(pos,number,BoxSize,W) #same as NGPW but double precision for number
# CICW_d(pos,number,BoxSize,W) #same as CICW but double precision for number

#### given a 3D density field, find value of it at particle positions ####
# CIC_interp(pos,density,BoxSize,dens)

#### from particle positions (2D or 3D) with radii to 2D density field ####
# voronoi_NGP_2D(density,pos,mass,volume,x_min,y_min,BoxSize,
#                          particles_per_cell,r_divisions)
# voronoi_RT_2D(density,pos,mass,radius,x_min,y_min,BoxSize, periodic,verbose)
# SPH_RT_2D(density,pos,mass,radius,x_min,y_min,axis_x,axis_y,
#           BoxSize,periodic,verbose)

# TO-DO: 2D computations are suboptimal for CIC,TSC and PCS as particles along
# the axis 2 are repeated 2,3 and 4 times, respectively
################################################################################
################################################################################

# This is the main function to use when performing the mass assignment
# pos --------> array containing the positions of the particles: 2D or 3D
# numbers ----> array containing the density field: 2D or 3D
# BoxSize ----> size of the simulation box
# MAS --------> mass assignment scheme: NGP, CIC, TSC or PCS
# W ----------> array containing the weights to be used: 1D array (optional)
# renormalize_2D ---> when computing the density field by reading multiple 
# subfiles, the normalization factor /2.0, /3.0, /4.0 should be added manually
# only at the end, otherwise the results will be incorrect!!
cpdef void MA(pos, number, BoxSize, MAS='CIC', W=None, verbose=False,
              renormalize_2D=True):

    #number of coordinates to work in 2D or 3D
    coord,coord_aux = pos.shape[1], number.ndim 

    # check that the number of dimensions match
    if coord!=coord_aux:
        print('pos have %d dimensions and the density %d!!!'%(coord,coord_aux))
        sys.exit()

    if verbose:
        if W is None:  print('\nUsing %s mass assignment scheme'%MAS)
        else:          print('\nUsing %s mass assignment scheme with weights'%MAS)
    start = time.time()
    if coord==3: 
        if   MAS=='NGP' and W is None:  NGP(pos,number,BoxSize)
        elif MAS=='CIC' and W is None:  CIC(pos,number,BoxSize)
        elif MAS=='TSC' and W is None:  TSC(pos,number,BoxSize)
        elif MAS=='PCS' and W is None:  PCS(pos,number,BoxSize)
        elif MAS=='NGP' and W is not None:  NGPW(pos,number,BoxSize,W)
        elif MAS=='CIC' and W is not None:  CICW(pos,number,BoxSize,W)
        elif MAS=='TSC' and W is not None:  TSCW(pos,number,BoxSize,W)
        elif MAS=='PCS' and W is not None:  PCSW(pos,number,BoxSize,W)
        else:
            print('option not valid!!!');  sys.exit()

    if coord==2:
        number2 = np.expand_dims(number,axis=2)
        if   MAS=='NGP' and W is None:  
            NGP(pos,number2,BoxSize)
        elif MAS=='CIC' and W is None:  
            CIC(pos,number2,BoxSize);
            if renormalize_2D:  number2 /= 2.0
        elif MAS=='TSC' and W is None:  
            TSC(pos,number2,BoxSize);  
            if renormalize_2D:  number2 /= 3.0
        elif MAS=='PCS' and W is None:  
            PCS(pos,number2,BoxSize);
            if renormalize_2D:  number2 /= 4.0
        elif MAS=='NGP' and W is not None:  
            NGPW(pos,number2,BoxSize,W)
        elif MAS=='CIC' and W is not None:  
            CICW(pos,number2,BoxSize,W);
            if renormalize_2D:  number2 /= 2.0
        elif MAS=='TSC' and W is not None:  
            TSCW(pos,number2,BoxSize,W);
            if renormalize_2D:  number2 /= 3.0
        elif MAS=='PCS' and W is not None:  
            PCSW(pos,number2,BoxSize,W); 
            if renormalize_2D:  number2 /= 4.0
        else:
            print('option not valid!!!');  sys.exit()
        number = number2[:,:,0]
    if verbose:
        print('Time taken = %.3f seconds\n'%(time.time()-start))
    

################################################################################
# This function computes the density field of a cubic distribution of particles
# pos ------> positions of the particles. Numpy array
# number ---> array with the density field. Numpy array (dims,dims,dims)
# BoxSize --> Size of the box
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True) 
cpdef void CIC(np.float32_t[:,:] pos, np.float32_t[:,:,:] number, float BoxSize):
        
    cdef int axis,dims,coord
    cdef long i,particles
    cdef float inv_cell_size,dist
    cdef float u[3]
    cdef float d[3]
    cdef int index_u[3]
    cdef int index_d[3]

    # find number of particles, the inverse of the cell size and dims
    particles = pos.shape[0];  coord = pos.shape[1];  dims = number.shape[0]
    inv_cell_size = dims/BoxSize
    
    # when computing things in 2D, use the index_ud[2]=0 plane
    for i in range(3):
        index_d[i] = 0;  index_u[i] = 0;  d[i] = 1.0;  u[i] = 1.0

    # do a loop over all particles
    for i in range(particles):

        # $: grid point, X: particle position
        # $.........$..X......$
        # ------------>         dist    (1.3)
        # --------->            index_d (1)
        # --------------------> index_u (2)
        #           --->        u       (0.3)
        #              -------> d       (0.7)
        for axis in range(coord):
            dist          = pos[i,axis]*inv_cell_size
            u[axis]       = dist - <int>dist
            d[axis]       = 1.0 - u[axis]
            index_d[axis] = (<int>dist)%dims
            index_u[axis] = index_d[axis] + 1
            index_u[axis] = index_u[axis]%dims #seems this is faster

        number[index_d[0],index_d[1],index_d[2]] += d[0]*d[1]*d[2]
        number[index_d[0],index_d[1],index_u[2]] += d[0]*d[1]*u[2]
        number[index_d[0],index_u[1],index_d[2]] += d[0]*u[1]*d[2]
        number[index_d[0],index_u[1],index_u[2]] += d[0]*u[1]*u[2]
        number[index_u[0],index_d[1],index_d[2]] += u[0]*d[1]*d[2]
        number[index_u[0],index_d[1],index_u[2]] += u[0]*d[1]*u[2]
        number[index_u[0],index_u[1],index_d[2]] += u[0]*u[1]*d[2]
        number[index_u[0],index_u[1],index_u[2]] += u[0]*u[1]*u[2]
################################################################################

################################################################################
# This function computes the density field of a cubic distribution of particles
# using weights
# pos ------> positions of the particles. Numpy array
# number ---> array with the density field. Numpy array (dims,dims,dims)
# BoxSize --> Size of the box
# W --------> weights of the particles
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void CICW(np.float32_t[:,:] pos, np.float32_t[:,:,:] number, float BoxSize,
               np.float32_t[:] W):

    cdef int axis,dims,coord
    cdef long i,particles
    cdef float inv_cell_size,dist
    cdef float u[3]
    cdef float d[3]
    cdef int index_d[3]
    cdef int index_u[3]
    
    # find number of particles, the inverse of the cell size and dims
    particles = pos.shape[0];  coord = pos.shape[1];  dims = number.shape[0]
    inv_cell_size = dims/BoxSize
    
    # when computing things in 2D, use the index_ud[2]=0 plane
    for i in range(3):
        index_d[i] = 0;  index_u[i] = 0;  d[i] = 1.0;  u[i] = 1.0

    # do a loop over all particles
    for i in range(particles):

        for axis in range(coord):
            dist          = pos[i,axis]*inv_cell_size
            u[axis]       = dist - <int>dist
            d[axis]       = 1.0 - u[axis]
            index_d[axis] = (<int>dist)%dims
            index_u[axis] = index_d[axis] + 1
            index_u[axis] = index_u[axis]%dims #seems this is faster

        number[index_d[0],index_d[1],index_d[2]] += d[0]*d[1]*d[2]*W[i]
        number[index_d[0],index_d[1],index_u[2]] += d[0]*d[1]*u[2]*W[i]
        number[index_d[0],index_u[1],index_d[2]] += d[0]*u[1]*d[2]*W[i]
        number[index_d[0],index_u[1],index_u[2]] += d[0]*u[1]*u[2]*W[i]
        number[index_u[0],index_d[1],index_d[2]] += u[0]*d[1]*d[2]*W[i]
        number[index_u[0],index_d[1],index_u[2]] += u[0]*d[1]*u[2]*W[i]
        number[index_u[0],index_u[1],index_d[2]] += u[0]*u[1]*d[2]*W[i]
        number[index_u[0],index_u[1],index_u[2]] += u[0]*u[1]*u[2]*W[i]
################################################################################

################################################################################
# This function computes the density field of a cubic distribution of particles
# using weights
# pos ------> positions of the particles. Numpy array
# number ---> array with the density field. Numpy array (dims,dims,dims)
# BoxSize --> Size of the box
# W --------> weights of the particles
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void CICW_d(np.float32_t[:,:] pos, np.float64_t[:,:,:] number, float BoxSize,
               np.float32_t[:] W):

    cdef int axis,dims,coord
    cdef long i,particles
    cdef float inv_cell_size,dist
    cdef float u[3]
    cdef float d[3]
    cdef int index_d[3]
    cdef int index_u[3]
    
    # find number of particles, the inverse of the cell size and dims
    particles = pos.shape[0];  coord = pos.shape[1];  dims = number.shape[0]
    inv_cell_size = dims/BoxSize
    
    # when computing things in 2D, use the index_ud[2]=0 plane
    for i in range(3):
        index_d[i] = 0;  index_u[i] = 0;  d[i] = 1.0;  u[i] = 1.0

    # do a loop over all particles
    for i in range(particles):

        for axis in range(coord):
            dist          = pos[i,axis]*inv_cell_size
            u[axis]       = dist - <int>dist
            d[axis]       = 1.0 - u[axis]
            index_d[axis] = (<int>dist)%dims
            index_u[axis] = index_d[axis] + 1
            index_u[axis] = index_u[axis]%dims #seems this is faster

        number[index_d[0],index_d[1],index_d[2]] += d[0]*d[1]*d[2]*W[i]
        number[index_d[0],index_d[1],index_u[2]] += d[0]*d[1]*u[2]*W[i]
        number[index_d[0],index_u[1],index_d[2]] += d[0]*u[1]*d[2]*W[i]
        number[index_d[0],index_u[1],index_u[2]] += d[0]*u[1]*u[2]*W[i]
        number[index_u[0],index_d[1],index_d[2]] += u[0]*d[1]*d[2]*W[i]
        number[index_u[0],index_d[1],index_u[2]] += u[0]*d[1]*u[2]*W[i]
        number[index_u[0],index_u[1],index_d[2]] += u[0]*u[1]*d[2]*W[i]
        number[index_u[0],index_u[1],index_u[2]] += u[0]*u[1]*u[2]*W[i]
################################################################################

################################################################################
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void NGP(np.float32_t[:,:] pos, np.float32_t[:,:,:] number, float BoxSize):

    cdef int axis,dims,coord
    cdef long i,particles
    cdef float inv_cell_size
    cdef int index[3]

    # find number of particles, the inverse of the cell size and dims
    particles = pos.shape[0];  coord = pos.shape[1];  dims = number.shape[0]
    inv_cell_size = dims/BoxSize

    # when computing things in 2D, use the index[2]=0 plane
    for i in range(3):  index[i] = 0

    # do a loop over all particles
    for i in range(particles):
        for axis in range(coord):
            index[axis] = <int>(pos[i,axis]*inv_cell_size + 0.5)
            index[axis] = (index[axis]+dims)%dims
        number[index[0],index[1],index[2]] += 1.0
################################################################################

################################################################################
# This function computes the density field of a cubic distribution of particles
# using weights
# pos ------> positions of the particles. Numpy array
# number ---> array with the density field. Numpy array (dims,dims,dims)
# BoxSize --> Size of the box
# W --------> weights of the particles
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void NGPW(np.float32_t[:,:] pos, np.float32_t[:,:,:] number, float BoxSize,
               np.float32_t[:] W):

    cdef int axis,dims,coord
    cdef long i,particles
    cdef float inv_cell_size
    cdef int index[3]

    # find number of particles, the inverse of the cell size and dims
    particles = pos.shape[0];  coord = pos.shape[1];  dims = number.shape[0]
    inv_cell_size = dims/BoxSize

    # when computing things in 2D, use the index[2]=0 plane
    for i in range(3):  index[i] = 0

    # do a loop over all particles
    for i in range(particles):
        for axis in range(coord):
            index[axis] = <int>(pos[i,axis]*inv_cell_size + 0.5)
            index[axis] = (index[axis]+dims)%dims
        number[index[0],index[1],index[2]] += W[i]
################################################################################

################################################################################
# This function computes the density field of a cubic distribution of particles
# using weights
# pos ------> positions of the particles. Numpy array
# number ---> array with the density field. Numpy array (dims,dims,dims)
# BoxSize --> Size of the box
# W --------> weights of the particles
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void NGPW_d(np.float32_t[:,:] pos, np.float64_t[:,:,:] number, 
                 float BoxSize, np.float32_t[:] W):

    cdef int axis,dims,coord
    cdef long i,particles
    cdef float inv_cell_size
    cdef int index[3]

    # find number of particles, the inverse of the cell size and dims
    particles = pos.shape[0];  coord = pos.shape[1];  dims = number.shape[0]
    inv_cell_size = dims/BoxSize

    # when computing things in 2D, use the index[2]=0 plane
    for i in range(3):  index[i] = 0

    # do a loop over all particles
    for i in range(particles):
        for axis in range(coord):
            index[axis] = <int>(pos[i,axis]*inv_cell_size + 0.5)
            index[axis] = (index[axis]+dims)%dims
        number[index[0],index[1],index[2]] += W[i]
################################################################################

################################################################################
# This function computes the density field of a cubic distribution of particles
# pos ------> positions of the particles. Numpy array
# number ---> array with the density field. Numpy array (dims,dims,dims)
# BoxSize --> Size of the box
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void TSC(np.float32_t[:,:] pos, np.float32_t[:,:,:] number, float BoxSize):

    cdef int axis, dims, minimum
    cdef int j, l, m, n, coord
    cdef long i, particles
    cdef float inv_cell_size, dist, diff
    cdef float C[3][3]
    cdef int index[3][3]

    # find number of particles, the inverse of the cell size and dims
    particles = pos.shape[0];  coord = pos.shape[1];  dims = number.shape[0]
    inv_cell_size = dims/BoxSize
    
    # define arrays: for 2D set we have C[2,:] = 1.0 and index[2,:] = 0
    for i in range(3):
        for j in range(3):
            C[i][j] = 1.0;  index[i][j] = 0
            
    # do a loop over all particles
    for i in range(particles):

        # do a loop over the axes of the particle
        for axis in range(coord):
            dist    = pos[i,axis]*inv_cell_size
            minimum = <int>floor(dist-1.5)
            for j in range(3): #only 3 cells/dimension can contribute
                index[axis][j] = (minimum+j+1+dims)%dims
                diff = fabs(minimum + j+1 - dist)
                if diff<0.5:    C[axis][j] = 0.75-diff*diff
                elif diff<1.5:  C[axis][j] = 0.5*(1.5-diff)*(1.5-diff)
                else:           C[axis][j] = 0.0

        for l in range(3):  
            for m in range(3):  
                for n in range(3): 
                    number[index[0][l],index[1][m],index[2][n]] += C[0][l]*C[1][m]*C[2][n]
################################################################################

################################################################################
# This function computes the density field of a cubic distribution of particles
# using weights
# pos ------> positions of the particles. Numpy array
# number ---> array with the density field. Numpy array (dims,dims,dims)
# BoxSize --> Size of the box
# W --------> weights of the particles
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void TSCW(np.float32_t[:,:] pos, np.float32_t[:,:,:] number, float BoxSize,
               np.float32_t[:] W):

    cdef int axis,dims,minimum,j,l,m,n,coord
    cdef long i,particles
    cdef float inv_cell_size,dist,diff
    cdef float C[3][3]
    cdef int index[3][3]

    # find number of particles, the inverse of the cell size and dims
    particles = pos.shape[0];  coord = pos.shape[1];  dims = number.shape[0]
    inv_cell_size = dims/BoxSize
    
    # define arrays: for 2D set we have C[2,:] = 1.0 and index[2,:] = 0
    for i in range(3):
        for j in range(3):
            C[i][j] = 1.0;  index[i][j] = 0
    
    # do a loop over all particles
    for i in range(particles):

        # do a loop over the three axes of the particle
        for axis in range(coord):
            dist    = pos[i,axis]*inv_cell_size
            minimum = <int>floor(dist-1.5)
            for j in range(3): #only 3 cells/dimension can contribute
                index[axis][j] = (minimum+ j+1+ dims)%dims
                diff = fabs(minimum+ j+1 - dist)
                if diff<0.5:    C[axis][j] = 0.75-diff*diff
                elif diff<1.5:  C[axis][j] = 0.5*(1.5-diff)*(1.5-diff)
                else:           C[axis][j] = 0.0

        for l in range(3):  
            for m in range(3):  
                for n in range(3): 
                    number[index[0][l],index[1][m],index[2][n]] += C[0][l]*C[1][m]*C[2][n]*W[i]
################################################################################

################################################################################
# This function computes the density field of a cubic distribution of particles
# pos ------> positions of the particles. Numpy array
# number ---> array with the density field. Numpy array (dims,dims,dims)
# BoxSize --> Size of the box
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void PCS(np.float32_t[:,:] pos, np.float32_t[:,:,:] number, float BoxSize):

    cdef int axis,dims,minimum,j,l,m,n,coord
    cdef long i,particles
    cdef float inv_cell_size,dist,diff
    cdef float C[3][4]
    cdef int index[3][4]

    # find number of particles, the inverse of the cell size and dims
    particles = pos.shape[0];  coord = pos.shape[1];  dims = number.shape[0]
    inv_cell_size = dims/BoxSize
        
    # define arrays: for 2D set we have C[2,:] = 1.0 and index[2,:] = 0
    for i in range(3):
        for j in range(4):
            C[i][j] = 1.0;  index[i][j] = 0

    # do a loop over all particles
    for i in range(particles):

        # do a loop over the three axes of the particle
        for axis in range(coord):
            dist    = pos[i,axis]*inv_cell_size
            minimum = <int>floor(dist-2.0)
            for j in range(4): #only 4 cells/dimension can contribute
                index[axis][j] = (minimum + j+1 + dims)%dims
                diff = fabs(minimum + j+1 - dist)
                if diff<1.0:    C[axis][j] = (4.0 - 6.0*diff*diff + 3.0*diff*diff*diff)/6.0
                elif diff<2.0:  C[axis][j] = (2.0 - diff)*(2.0 - diff)*(2.0 - diff)/6.0
                else:           C[axis][j] = 0.0

        for l in range(4):  
            for m in range(4):  
                for n in range(4): 
                    number[index[0][l],index[1][m],index[2][n]] += C[0][l]*C[1][m]*C[2][n]
################################################################################

################################################################################
# This function computes the density field of a cubic distribution of particles
# using weights
# pos ------> positions of the particles. Numpy array
# number ---> array with the density field. Numpy array (dims,dims,dims)
# BoxSize --> Size of the box
# W --------> weights of the particles
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void PCSW(np.float32_t[:,:] pos, np.float32_t[:,:,:] number, float BoxSize,
               np.float32_t[:] W):

    cdef int axis,dims,minimum,j,l,m,n,coord
    cdef long i,particles
    cdef float inv_cell_size,dist,diff
    cdef float C[3][4]
    cdef int index[3][4]

    # find number of particles, the inverse of the cell size and dims
    particles = pos.shape[0];  coord = pos.shape[1];  dims = number.shape[0]
    inv_cell_size = dims/BoxSize

    # define arrays: for 2D set we have C[2,:] = 1.0 and index[2,:] = 0
    for i in range(3):
        for j in range(4):
            C[i][j] = 1.0;  index[i][j] = 0
    
    # do a loop over all particles
    for i in range(particles):

        # do a loop over the three axes of the particle
        for axis in range(coord):
            dist    = pos[i,axis]*inv_cell_size
            minimum = <int>floor(dist-2.0)
            for j in range(4): #only 4 cells/dimension can contribute
                index[axis][j] = (minimum + j+1 + dims)%dims
                diff = fabs(minimum + j+1 - dist)
                if diff<1.0:    C[axis][j] = (4.0 - 6.0*diff*diff + 3.0*diff*diff*diff)/6.0
                elif diff<2.0:  C[axis][j] = (2.0 - diff)*(2.0 - diff)*(2.0 - diff)/6.0
                else:           C[axis][j] = 0.0

        for l in range(4):
            for m in range(4):
                for n in range(4): 
                    number[index[0][l],index[1][m],index[2][n]] += C[0][l]*C[1][m]*C[2][n]*W[i]
################################################################################

################################################################################
# This function takes a 3D grid called density. The routine finds the CIC 
# interpolated value of the grid onto the positions input as pos
# density --> 3D array with containing the density field
# BoxSize --> Size of the box
# pos ------> positions where the density field will be interpolated
# den ------> array with the interpolated density field at pos
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void CIC_interp(np.ndarray[np.float32_t,ndim=3] density, float BoxSize,
                      np.ndarray[np.float32_t,ndim=2] pos,
                      np.ndarray[np.float32_t,ndim=1] den):

    cdef int axis,dims
    cdef long i,particles
    cdef float inv_cell_size,dist
    cdef float u[3]
    cdef float d[3]
    cdef int index_u[3]
    cdef int index_d[3]
    
    # find number of particles, the inverse of the cell size and dims
    particles = pos.shape[0];  dims = density.shape[0]
    inv_cell_size = dims/BoxSize

    # do a loop over all particles
    for i in range(particles):

        # $: grid point, X: particle position
        # $.........$..X......$
        # ------------>         dist    (1.3)
        # --------->            index_d (1)
        # --------------------> index_u (2)
        #           --->        u       (0.3)
        #              -------> d       (0.7)
        for axis in range(3):
            dist          = pos[i,axis]*inv_cell_size
            u[axis]       = dist - <int>dist
            d[axis]       = 1.0 - u[axis]
            index_d[axis] = (<int>dist)%dims
            index_u[axis] = index_d[axis] + 1
            index_u[axis] = index_u[axis]%dims #seems this is faster

        den[i] = density[index_d[0],index_d[1],index_d[2]]*d[0]*d[1]*d[2]+\
                 density[index_d[0],index_d[1],index_u[2]]*d[0]*d[1]*u[2]+\
                 density[index_d[0],index_u[1],index_d[2]]*d[0]*u[1]*d[2]+\
                 density[index_d[0],index_u[1],index_u[2]]*d[0]*u[1]*u[2]+\
                 density[index_u[0],index_d[1],index_d[2]]*u[0]*d[1]*d[2]+\
                 density[index_u[0],index_d[1],index_u[2]]*u[0]*d[1]*u[2]+\
                 density[index_u[0],index_u[1],index_d[2]]*u[0]*u[1]*d[2]+\
                 density[index_u[0],index_u[1],index_u[2]]*u[0]*u[1]*u[2]
################################################################################


# This routine computes the 2D density field from a set of voronoi cells that
# have masses and volumes. It considers each particle as a sphere and split it 
# into particles_per_cell tracers. 
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void voronoi_NGP_2D(np.float64_t[:,:] field, np.float32_t[:,:] pos,
                          np.float32_t[:] mass, np.float32_t[:] radii,
                          float x_min, float y_min, float BoxSize,
                          long tracers, int r_divisions, periodic):

    cdef float pi = np.pi
    cdef long i, j, k, particles, grid, count
    cdef double dtheta, angle, R, R1, R2, area, length, V_sphere, norm
    cdef double R_cell, W_cell, x_cell, y_cell
    cdef np.float32_t[:,:] pos_tracer
    cdef np.float32_t[:] w_tracer
    cdef double radius, x, y, w, inv_cell_size
    cdef int index_x, index_y, index_xp, index_xm, index_yp, index_ym
    cdef int theta_divisions

    # verbose
    print('Calculating projected mass of the voronoi tracers...')
    start = time.time()

    # find the number of particles to analyze and inv_cell_size
    particles     = pos.shape[0]
    grid          = field.shape[0]
    inv_cell_size = grid*1.0/BoxSize

    # compute the number of particles in each shell and the angle between them
    theta_divisions = tracers//r_divisions
    dtheta          = 2.0*pi/theta_divisions
    V_sphere        = 4.0*pi*1.0**3/3.0

    # define the arrays with the properties of the tracers; positions and weights
    pos_tracer = np.zeros((tracers,2), dtype=np.float32)
    w_tracer   = np.zeros(tracers,     dtype=np.float32)

    # define and fill the array containing pos_tracer
    count = 0
    for i in range(r_divisions):
        R1, R2 = i*1.0/r_divisions, (i+1.0)/r_divisions
        R = 0.5*(R1 + R2)
        area = pi*(R2**2 - R1**2)/theta_divisions
        length = 2.0*sqrt(1.0**2 - R**2)
        for j in range(theta_divisions):
            angle = 2.0*pi*(j + 0.5)/theta_divisions
            pos_tracer[count,0] = R*cos(angle)
            pos_tracer[count,1] = R*sin(angle)
            w_tracer[count]     = area*length/V_sphere
            count += 1

    # normalize weights of tracers to force them to sum 1
    norm = np.sum(w_tracer, dtype=np.float64)
    for i in range(tracers):
        w_tracer[i] = w_tracer[i]/norm
        
    if periodic:
        
        # do a loop over all particles
        for i in range(particles):

            R_cell = radii[i]
            W_cell = mass[i]
            x_cell = pos[i,0]
            y_cell = pos[i,1]

            # see if we need to split the particle into tracers or not
            index_xm = <int>round((x_cell-R_cell-x_min)*inv_cell_size - 0.5)
            index_xp = <int>round((x_cell+R_cell-x_min)*inv_cell_size - 0.5)
            index_ym = <int>round((y_cell-R_cell-y_min)*inv_cell_size - 0.5)
            index_yp = <int>round((y_cell+R_cell-y_min)*inv_cell_size - 0.5)

            if (index_xm==index_xp) and (index_ym==index_yp):
                index_x = (index_xm + grid)%grid
                index_y = (index_ym + grid)%grid
                field[index_x, index_y] += W_cell
                
            else:

                # do a loop over all tracers
                for j in range(tracers):

                    x = x_cell + R_cell*pos_tracer[j,0]
                    y = y_cell + R_cell*pos_tracer[j,1]
                    w = W_cell*w_tracer[j]

                    index_x = <int>round((x-x_min)*inv_cell_size - 0.5)
                    index_y = <int>round((y-y_min)*inv_cell_size - 0.5)
                    index_x = (index_x + grid)%grid
                    index_y = (index_y + grid)%grid
                    
                    field[index_x, index_y] += w

    else:

        # do a loop over all particles
        for i in range(particles):
        
            R_cell = radii[i]
            W_cell = mass[i]
            x_cell = pos[i,0]
            y_cell = pos[i,1]

            # do a loop over all tracers
            for j in range(tracers):

                x = x_cell + R_cell*pos_tracer[j,0]
                y = y_cell + R_cell*pos_tracer[j,1]
                w = W_cell*w_tracer[j]

                index_x = <int>round((x-x_min)*inv_cell_size - 0.5)
                index_y = <int>round((y-y_min)*inv_cell_size - 0.5)

                if (index_x<0) or (index_x>=grid):  continue
                if (index_y<0) or (index_y>=grid):  continue
                
                field[index_x, index_y] += w

    print('Time taken = %.3f s'%(time.time()-start))


# This routine computes the 2D density field from a set of voronoi cells that
# have masses and volumes. It considers each particle as a sphere and split it 
# into particles_per_cell tracers. 
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void projected_voronoi(np.float64_t[:,:] field, np.float32_t[:,:] pos,
                             np.float32_t[:] mass, np.float32_t[:] radii,
                             float x_min, float y_min, float BoxSize,
                             long tracers, int r_divisions, periodic, verbose=True):

    cdef double pi = np.pi
    cdef long i, j, k, particles, grid, count
    cdef double angle, R, R1, R2, area, length, V_sphere, norm
    cdef double R_cell, W_cell, x_cell, y_cell
    cdef double x, y, w, inv_cell_size
    cdef np.float64_t[:,:] pos_tracer
    cdef np.float64_t[:] w_tracer
    cdef int index_x, index_y, index_xp, index_xm, index_yp, index_ym
    cdef int theta_divisions

    # verbose
    if verbose:  print('Calculating projected mass of the voronoi tracers...')
    start = time.time()

    # find the number of particles to analyze and inv_cell_size
    particles     = pos.shape[0]
    grid          = field.shape[0]
    inv_cell_size = grid*1.0/BoxSize

    # compute the number of particles in each shell and the angle between them
    theta_divisions = tracers//r_divisions
    V_sphere        = 4.0*pi*1.0**3/3.0

    # define the arrays with the properties of the tracers; positions and weights
    pos_tracer = np.zeros((tracers,2), dtype=np.float64)
    w_tracer   = np.zeros(tracers,     dtype=np.float64)

    # define and fill the array containing pos_tracer
    count = 0
    for i in range(r_divisions)[::-1]:
        R1, R2 = i*1.0/r_divisions, (i+1.0)/r_divisions
        R = 0.5*(R1 + R2)
        area = pi*(R2**2 - R1**2)/theta_divisions
        length = 2.0*sqrt(1.0**2 - R**2)
        for j in range(theta_divisions):
            angle = 2.0*pi*(j + 0.5)/theta_divisions
            pos_tracer[count,0] = R*cos(angle)
            pos_tracer[count,1] = R*sin(angle)
            w_tracer[count]     = area*length/V_sphere
            count += 1

    # normalize weights of tracers to force them to sum 1
    norm = np.sum(w_tracer, dtype=np.float64)
    for i in range(tracers):
        w_tracer[i] = w_tracer[i]/norm
        
    # when using periodic boundary conditions
    if periodic:
        
        # do a loop over all particles
        for i in range(particles):

            R_cell = radii[i]
            W_cell = mass[i]
            x_cell = pos[i,0]
            y_cell = pos[i,1]

            # do a loop over the different radii
            count = 0
            for j in range(r_divisions)[::-1]:
                R1, R2 = j*1.0/r_divisions, (j+1.0)/r_divisions
                R = 0.5*(R1 + R2)*R_cell

                # see if we need to split the particle into tracers or not
                index_xm = <int>round((x_cell-R-x_min)*inv_cell_size - 0.5)
                index_xp = <int>round((x_cell+R-x_min)*inv_cell_size - 0.5)
                index_ym = <int>round((y_cell-R-y_min)*inv_cell_size - 0.5)
                index_yp = <int>round((y_cell+R-y_min)*inv_cell_size - 0.5)

                # if particles in the shell are all within the same pixel
                if (index_xm==index_xp) and (index_ym==index_yp):
                    index_x = (index_xm + grid)%grid
                    index_y = (index_ym + grid)%grid
                    for k in range(count, tracers):
                        field[index_x, index_y] += W_cell*w_tracer[count]
                        count += 1
                    break
                    
                else:

                    # do a loop over the particles in the shell
                    for k in range(theta_divisions):

                        x = x_cell + R_cell*pos_tracer[count,0]
                        y = y_cell + R_cell*pos_tracer[count,1]
                        w = W_cell*w_tracer[count]

                        index_x = <int>round((x-x_min)*inv_cell_size - 0.5)
                        index_y = <int>round((y-y_min)*inv_cell_size - 0.5)
                        index_x = (index_x + grid)%grid
                        index_y = (index_y + grid)%grid
                        
                        field[index_x, index_y] += w
                        count += 1

    # if periodic boundary conditions do not apply
    else:

        # do a loop over all particles
        for i in range(particles):
        
            R_cell = radii[i]
            W_cell = mass[i]
            x_cell = pos[i,0]
            y_cell = pos[i,1]

            # do a loop over the different radii
            count = 0
            for j in range(r_divisions)[::-1]:
                R1, R2 = j*1.0/r_divisions, (j+1.0)/r_divisions
                R = 0.5*(R1 + R2)*R_cell

                # see if we need to split the particle into tracers or not
                index_xm = <int>round((x_cell-R-x_min)*inv_cell_size - 0.5)
                index_xp = <int>round((x_cell+R-x_min)*inv_cell_size - 0.5)
                index_ym = <int>round((y_cell-R-y_min)*inv_cell_size - 0.5)
                index_yp = <int>round((y_cell+R-y_min)*inv_cell_size - 0.5)

                # if particles in the shell are all within the same pixel
                if (index_xm==index_xp) and (index_ym==index_yp):

                    if (index_xm<0) or (index_xm>=grid):  continue
                    if (index_ym<0) or (index_ym>=grid):  continue
                    index_x = (index_xm + grid)%grid
                    index_y = (index_ym + grid)%grid
                    for k in range(count, tracers):
                        field[index_x, index_y] += W_cell*w_tracer[count]
                        count += 1
                    break
                    
                else:

                    # do a loop over the particles in the shell
                    for k in range(theta_divisions):

                        x = x_cell + R_cell*pos_tracer[count,0]
                        y = y_cell + R_cell*pos_tracer[count,1]
                        w = W_cell*w_tracer[count]
                        count += 1

                        index_x = <int>round((x-x_min)*inv_cell_size - 0.5)
                        index_y = <int>round((y-y_min)*inv_cell_size - 0.5)

                        if (index_x<0) or (index_x>=grid):  continue
                        if (index_y<0) or (index_y>=grid):  continue
                        
                        field[index_x, index_y] += w
    
    if verbose:  print('Time taken = %.3f s'%(time.time()-start))



############################################################################### 
# This routine computes the 2D density field from a set of voronoi cells that
# have masses and radii assuming they represent uniform spheres. A cell that
# intersects with a cell will increase its value by the column density of the 
# cell along the sphere. It works with and without boundary conditions.
# The density array contains the column densities in units of Msun/(Mpc/h)^2 
# if positions are in Mpc/h and masses in Msun/h
# density ----------------> array hosting the column density field
# pos --------------------> array with positions of the particles 3D or 2D
# mass -------------------> array with the masses of the particles
# radius -----------------> array with the SPH radii of the particles
# x_min ------------------> minimum coordinate along first axis 
# y_min ------------------> minimum coordinate along second axis 
# axis_x -----------------> which component put along axis x (0-x, 1-y 2-z)
# axis_y -----------------> which component put along axis y (0-x, 1-y 2-z)
# BoxSize ----------------> size of the region
# periodic ---------------> whether there are boundary conditions (True/False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void voronoi_RT_2D(double[:,::1] density, float[:,::1] pos,
                         float[::1] mass, float[::1] radius,
                         float x_min, float y_min, int axis_x, int axis_y,
                         float BoxSize, periodic, verbose=True):

    start = time.time()
    cdef long particles, i
    cdef int dims, index_x, index_y, index_R, ii, jj, i_cell, j_cell
    cdef float x, y, rho, pi, cell_size, inv_cell_size, radius2
    cdef float dist2, dist2_x

    if verbose:  print('Computing column densities of the particles...')

    # find the number of particles and the dimensions of the grid
    particles = pos.shape[0]
    dims      = density.shape[0]
    pi        = np.pi

    # define cell size and the inverse of the cell size
    cell_size     = BoxSize*1.0/dims
    inv_cell_size = dims*1.0/BoxSize

    # if boundary conditions
    if periodic:
        for i in range(particles):

            # find the density of the particle and the square of its radius
            rho     = 3.0*mass[i]/(4.0*pi*radius[i]**3) #h^2 Msun/Mpc^3
            radius2 = radius[i]**2                      #(Mpc/h)^2

            # find cell where the particle center is and its radius in cell units
            index_x = <int>round((pos[i,axis_x]-x_min)*inv_cell_size - 0.5)
            index_y = <int>round((pos[i,axis_y]-y_min)*inv_cell_size - 0.5)
            index_R = <int>round(radius[i]*inv_cell_size - 0.5) + 1

            # do a loop over the cells that contribute in the x-direction
            for ii in range(-index_R, index_R+1):
                x       = (index_x + ii)*cell_size + x_min
                i_cell  = ((index_x + ii + dims)%dims)
                dist2_x = (x-pos[i,axis_x])**2 

                # do a loop over the cells that contribute in the y-direction
                for jj in range(-index_R, index_R+1):
                    y      = (index_y + jj)*cell_size + y_min
                    j_cell = ((index_y + jj + dims)%dims)

                    dist2 = dist2_x + (y-pos[i,axis_y])**2

                    if dist2<radius2:
                        density[i_cell,j_cell] += 2.0*rho*sqrt(radius2 - dist2)
    
    # if no periodic boundary conditions
    else:
        for i in range(particles):

            # find the density of the particle and the square of its radius
            rho     = 3.0*mass[i]/(4.0*pi*radius[i]**3) #h^2 Msun/Mpc^3
            radius2 = radius[i]**2                      #(Mpc/h)^2

            # find cell where the particle center is and its radius in cell units
            index_x = <int>round((pos[i,axis_x]-x_min)*inv_cell_size - 0.5)
            index_y = <int>round((pos[i,axis_y]-y_min)*inv_cell_size - 0.5)
            index_R = <int>round(radius[i]*inv_cell_size - 0.5) + 1

            # do a loop over the cells that contribute in the x-direction
            for ii in range(-index_R, index_R+1):
                i_cell = index_x + ii
                if i_cell>=0 and i_cell<dims:
                    x = i_cell*cell_size + x_min
                    dist2_x = (x-pos[i,axis_x])**2 
                else:  continue
                    
                # do a loop over the cells that contribute in the y-direction
                for jj in range(-index_R, index_R+1):
                    j_cell = index_y + jj
                    if j_cell>=0 and j_cell<dims:
                        y = j_cell*cell_size + y_min
                    else: continue

                    dist2 = dist2_x + (y-pos[i,axis_y])**2

                    if dist2<radius2:
                        density[i_cell,j_cell] += 2.0*rho*sqrt(radius2 - dist2)

    if verbose:  print('Time taken = %.2f seconds'%(time.time()-start))


############################################################################### 
# This function generates 2N+1 points in the surface of a sphere of radius 1
# uniformly distributed. See https://arxiv.org/pdf/0912.4540.pdf for details
def sphere_points(N):

    points = np.empty((2*N+1, 3), dtype=np.float32)
    
    i   = np.arange(-N, N+1, 1)
    lat = np.arcsin(2.0*i/(2.0*N+1))
    lon = 2.0*np.pi*i*2.0/(1.0+np.sqrt(5.0))
    
    points[:,0] = np.cos(lat)*np.cos(lon)
    points[:,1] = np.cos(lat)*np.sin(lon)
    points[:,2] = np.sin(lat)
    
    return points

# The standard SPH kernel is: u = r/R, where R is the SPH radius
#                  [ 1-6u^2 + 6u^3  if 0<u<0.5
# W(r) = 8/(pi*R^3)[ 2(1-u)^3       if 0.5<u<1 
#                  [ 0              otherwise
# This function returns 4*pi int_0^r r^2 W(r) dr
# We use this function to split the sphere into shells of equal weight
def sph_kernel_volume(u):
    if u<0.5:    return 32.0/15.0*u**3*(5.0 - 18.0*u**2 + 15.0*u**3)
    elif u<=1.0:  return -1.0/15.0 - 64.0*(u**6/6.0 - 3.0/5.0*u**5 + 3.0/4.0*u**4 -u**3/3.0)
    else:        return 0.0

# This is just a vectorization of the above function
def vsph_kernel_volume(u):
    func = np.vectorize(sph_kernel_volume)
    return func(u)

# This routine computes the density field of a set of particles taking into
# account their physical extent given by its SPH radius
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void SPH_NGP(float[:,:,::1] density, float[:,::1] pos,  
                   float[::1] radius, int r_bins, int part_in_shell,
                   float BoxSize, verbose=True):

    cdef int dims, index_x, index_y, index_z
    cdef long i, j, num, particles, points_sph_sphere
    cdef float R, inv_cell_size, X, Y, Z
    cdef float[::1] r_values
    cdef float[:,::1] points, sph_points

    if verbose:  print('Findind density field using SPH radii of particles...')

    # determine the mean radii of the radial shells
    radial_bins = np.linspace(0, 1, r_bins+1)
    u_array  = np.linspace(0, 1, 1000)
    V_array  = vsph_kernel_volume(u_array)
    r_values = (np.interp(radial_bins, V_array, u_array)).astype(np.float32)
    for i in range(r_bins):
        r_values[i] = 0.5*(r_values[i]+r_values[i+1])

    # generate points uniformly distributed in a sphere
    points = sphere_points(part_in_shell)

    # define the array containing the points in the SPH sphere
    points_sph_sphere = r_bins*points.shape[0]
    sph_points = np.empty((points_sph_sphere, 3), dtype=np.float32)

    # find the position of all particles in the normalized SPH sphere
    num = 0
    for i in range(r_bins):
        for j in range(points.shape[0]):
            sph_points[num,0] = r_values[i]*points[j,0]
            sph_points[num,1] = r_values[i]*points[j,1]
            sph_points[num,2] = r_values[i]*points[j,2]
            num += 1

    # find the total number of particles
    particles     = pos.shape[0]
    dims          = density.shape[0]
    inv_cell_size = dims*1.0/BoxSize

    # do a loop over all particles
    for i in range(particles):
        X = pos[i,0]*inv_cell_size;  Y = pos[i,1]*inv_cell_size;  
        Z = pos[i,2]*inv_cell_size;  R = radius[i]*inv_cell_size

        for j in range(points_sph_sphere):
            index_x = <int>round(-0.5 + (X + R*sph_points[j,0]))
            index_y = <int>round(-0.5 + (Y + R*sph_points[j,1]))
            index_z = <int>round(-0.5 + (Z + R*sph_points[j,2]))
            index_x = (index_x+dims)%dims
            index_y = (index_y+dims)%dims
            index_z = (index_z+dims)%dims

            density[index_x, index_y, index_z] += 1.0


# This routine computes the density field of a set of particles taking into
# account their physical extent given by its SPH radius. The contribution
# of each particle is weighted by W
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void SPH_NGPW(float[:,:,::1] density, float[:,::1] pos,
                    float[::1] radius, float[::1] W, int r_bins,
                    int part_in_shell, float BoxSize, verbose=True):

    cdef int dims, index_x, index_y, index_z
    cdef long i, j, num, particles, points_sph_sphere
    cdef float R, inv_cell_size, X, Y, Z
    cdef float[::1] r_values
    cdef float[:,::1] points, sph_points

    if verbose:  print('Findind density field using SPH radii of particles...')

    # determine the mean radii of the radial shells
    radial_bins = np.linspace(0, 1, r_bins+1)
    u_array  = np.linspace(0, 1, 1000)
    V_array  = vsph_kernel_volume(u_array)
    r_values = (np.interp(radial_bins, V_array, u_array)).astype(np.float32)
    for i in range(r_bins):
        r_values[i] = 0.5*(r_values[i]+r_values[i+1])

    # generate points uniformly distributed in a sphere
    points = sphere_points(part_in_shell)

    # define the array containing the points in the SPH sphere
    points_sph_sphere = r_bins*points.shape[0]
    sph_points = np.empty((points_sph_sphere, 3), dtype=np.float32)

    # find the position of all particles in the normalized SPH sphere
    num = 0
    for i in range(r_bins):
        for j in range(points.shape[0]):
            sph_points[num,0] = r_values[i]*points[j,0]
            sph_points[num,1] = r_values[i]*points[j,1]
            sph_points[num,2] = r_values[i]*points[j,2]
            num += 1

    # find the total number of particles
    particles     = pos.shape[0]
    dims          = density.shape[0]
    inv_cell_size = dims*1.0/BoxSize

    # do a loop over all particles
    for i in range(particles):
        X = pos[i,0]*inv_cell_size;  Y = pos[i,1]*inv_cell_size;  
        Z = pos[i,2]*inv_cell_size;  R = radius[i]*inv_cell_size
        weight = W[i]

        for j in range(points_sph_sphere):
            index_x = <int>round(-0.5 + (X + R*sph_points[j,0]))
            index_y = <int>round(-0.5 + (Y + R*sph_points[j,1]))
            index_z = <int>round(-0.5 + (Z + R*sph_points[j,2]))
            index_x = (index_x+dims)%dims
            index_y = (index_y+dims)%dims
            index_z = (index_z+dims)%dims

            density[index_x, index_y, index_z] += W[i]
################################################################################

################################################################################
# This is the SPH kernel we are using 
def kernel_SPH(r,R):

    u = r/R
    prefact = 8.0/(np.pi*R**3)
    if u<0.5:     return prefact*(1.0 - 6.0*u*u + 6.0*u*u*u)
    elif u<=1.0:  return prefact*2.0*(1.0 - u)**3
    else:  return 0.0

def integrand(x,b2):
    r = sqrt(b2 + x*x)
    return kernel_SPH(r,1.0)

# This function computes the integral of the SPH kernel
# int_0^lmax W(r) dl, where b^2 + l^2 = r^2. b is the impact parameter
def NHI_table(bins):

    # arrays with impact parameter^2 and the column densities
    b2s = np.linspace(0, 1, bins, dtype=np.float64)
    NHI = np.zeros(bins,          dtype=np.float64)

    for i,b2 in enumerate(b2s):
        if b2==1.0:  continue

        lmax = sqrt(1.0 - b2)
        I,dI = SI.quad(integrand, 0.0, lmax, 
            args=(b2,), epsabs=1e-12, epsrel=1e-12)
        NHI[i] = 2.0*I

    return b2s, NHI

# This routine computes the 2D density field from a set of particles with 
# radii (e.g. SPH) and masses. A cell that intersects with a particle will 
# increase its value by the column density of the cell along the sphere. 
# The density array contains the column densities in units of Msun/(Mpc/h)^2 
# if positions are in Mpc/h and masses in Msun/h
# density ----------------> array hosting the column density field
# pos --------------------> array with positions of the particles 3D or 2D
# mass -------------------> array with the masses of the particles
# radius -----------------> array with the SPH radii of the particles
# x_min ------------------> minimum coordinate along first axis 
# y_min ------------------> minimum coordinate along second axis 
# axis_x -----------------> which component put along axis x (0-x, 1-y 2-z)
# axis_y -----------------> which component put along axis y (0-x, 1-y 2-z)
# BoxSize ----------------> size of the region
# periodic ---------------> whether there are boundary conditions (True/False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef void SPH_RT_2D(double[:,::1] density, float[:,::1] pos,
                     float[::1] mass, float[::1] radius,
                     float x_min, float y_min, int axis_x, int axis_y,
                     float BoxSize, periodic, verbose=True):

    start = time.time()
    cdef long particles, i, num, bins = 1000
    cdef int dims, index_x, index_y, index_R, ii, jj, i_cell, j_cell
    cdef float x, y, pi, cell_size, inv_cell_size, radius2
    cdef float dist2, dist2_x, mass_part
    cdef double[::1] b2, NHI

    if verbose:  print('Computing column densities of the particles')

    # find the number of particles and the dimensions of the grid
    particles = pos.shape[0]
    dims      = density.shape[0]
    pi        = np.pi

    # define cell size and the inverse of the cell size
    cell_size     = BoxSize*1.0/dims
    inv_cell_size = dims*1.0/BoxSize

    # compute the normalized column density for normalized radii^2
    b2, NHI = NHI_table(bins)

    # periodic boundary conditions
    if periodic:
        for i in range(particles):

            # find the particle mass and the square of its radius
            radius2   = radius[i]**2 #(Mpc/h)^2
            mass_part = mass[i]

            # find cell where the particle center is and its radius in cell units
            index_x = <int>round((pos[i,axis_x]-x_min)*inv_cell_size - 0.5)
            index_y = <int>round((pos[i,axis_y]-y_min)*inv_cell_size - 0.5)
            index_R = <int>round(radius[i]*inv_cell_size - 0.5) + 1

            # do a loop over the cells that contribute in the x-direction
            for ii in range(-index_R, index_R+1):
                x       = (index_x + ii)*cell_size + x_min
                i_cell  = ((index_x + ii + dims)%dims)
                dist2_x = (x-pos[i,axis_x])**2 

                # do a loop over the cells that contribute in the y-direction
                for jj in range(-index_R, index_R+1):
                    y      = (index_y + jj)*cell_size + y_min
                    j_cell = ((index_y + jj + dims)%dims)

                    dist2 = dist2_x + (y-pos[i,axis_y])**2

                    if dist2<radius2:
                        num = <int>(dist2/radius2)*bins
                        density[i_cell,j_cell] += (mass_part*NHI[num])

    # no periodic boundary conditions
    else:
        for i in range(particles):

            # find the particle mass and the square of its radius
            radius2   = radius[i]**2 #(Mpc/h)^2
            mass_part = mass[i]

            # find cell where the particle center is and its radius in cell units
            index_x = <int>round((pos[i,axis_x]-x_min)*inv_cell_size - 0.5)
            index_y = <int>round((pos[i,axis_y]-y_min)*inv_cell_size - 0.5)
            index_R = <int>round(radius[i]*inv_cell_size - 0.5) + 1

            # do a loop over the cells that contribute in the x-direction
            for ii in range(-index_R, index_R+1):

                i_cell = index_x + ii
                if i_cell>=0 and i_cell<dims:
                    x = i_cell*cell_size + x_min
                else:  continue
                dist2_x = (x-pos[i,axis_x])**2 

                # do a loop over the cells that contribute in the y-direction
                for jj in range(-index_R, index_R+1):
                    
                    j_cell = index_y + jj
                    if j_cell>=0 and j_cell<dims:
                        y = j_cell*cell_size + y_min
                    else: continue

                    dist2 = dist2_x + (y-pos[i,axis_y])**2

                    if dist2<radius2:
                        num = <int>(dist2/radius2)*bins
                        density[i_cell,j_cell] += (mass_part*NHI[num])
            
    if verbose:  print('Time taken = %.2f seconds'%(time.time()-start))




##################### MAS_c (openmp) functions ########################

############## NGP #################
cpdef void NGPc3D(FLOAT[:,::1] pos, FLOAT[:,:,::1] number, FLOAT BoxSize,
                  long threads):
    MASC.NGP(&pos[0,0], &number[0,0,0], NULL, pos.shape[0], number.shape[0],
             pos.shape[1], BoxSize, threads)
    
cpdef void NGPWc3D(FLOAT[:,::1] pos, FLOAT[:,:,::1] number, FLOAT[::1] W,
                   FLOAT BoxSize, long threads):
    MASC.NGP(&pos[0,0], &number[0,0,0], &W[0], pos.shape[0], number.shape[0], 
             pos.shape[1], BoxSize, threads)

cpdef void NGPc2D(FLOAT[:,::1] pos, FLOAT[:,::1] number, FLOAT BoxSize,
                  long threads):
    MASC.NGP(&pos[0,0], &number[0,0], NULL, pos.shape[0], number.shape[0],
             pos.shape[1], BoxSize, threads)

cpdef void NGPWc2D(FLOAT[:,::1] pos, FLOAT[:,::1] number, FLOAT[::1] W,
                   FLOAT BoxSize, long threads):
    MASC.NGP(&pos[0,0], &number[0,0], &W[0], pos.shape[0], number.shape[0],
             pos.shape[1], BoxSize, threads)
####################################

############## CIC #################
cpdef void CICc3D(FLOAT[:,::1] pos, FLOAT[:,:,::1] number, FLOAT BoxSize,
                  long threads):
    MASC.CIC(&pos[0,0], &number[0,0,0], NULL, pos.shape[0], number.shape[0],
             pos.shape[1], BoxSize, threads)
    
cpdef void CICWc3D(FLOAT[:,::1] pos, FLOAT[:,:,::1] number, FLOAT[::1] W,
                   FLOAT BoxSize, long threads):
    MASC.CIC(&pos[0,0], &number[0,0,0], &W[0], pos.shape[0], number.shape[0],
             pos.shape[1], BoxSize, threads)

cpdef void CICc2D(FLOAT[:,::1] pos, FLOAT[:,::1] number, FLOAT BoxSize,
                  long threads):
    MASC.CIC(&pos[0,0], &number[0,0], NULL, pos.shape[0], number.shape[0],
             pos.shape[1], BoxSize, threads)

cpdef void CICWc2D(FLOAT[:,::1] pos, FLOAT[:,::1] number, FLOAT[::1] W,
                   FLOAT BoxSize, long threads):
    MASC.CIC(&pos[0,0], &number[0,0], &W[0], pos.shape[0], number.shape[0],
             pos.shape[1], BoxSize, threads)
####################################

############## TSC #################
cpdef void TSCc3D(FLOAT[:,::1] pos, FLOAT[:,:,::1] number, FLOAT BoxSize,
                  long threads):
    MASC.TSC(&pos[0,0], &number[0,0,0], NULL, pos.shape[0], number.shape[0],
             pos.shape[1], BoxSize, threads)

cpdef void TSCWc3D(FLOAT[:,::1] pos, FLOAT[:,:,::1] number, FLOAT[::1] W,
                   FLOAT BoxSize, long threads):
    MASC.TSC(&pos[0,0], &number[0,0,0], &W[0], pos.shape[0], number.shape[0],
             pos.shape[1], BoxSize, threads)

cpdef void TSCc2D(FLOAT[:,::1] pos, FLOAT[:,::1] number, FLOAT BoxSize,
                  long threads):
    MASC.TSC(&pos[0,0], &number[0,0], NULL, pos.shape[0], number.shape[0],
             pos.shape[1], BoxSize, threads)

cpdef void TSCWc2D(FLOAT[:,::1] pos, FLOAT[:,::1] number, FLOAT[::1] W,
                   FLOAT BoxSize, long threads):
    MASC.TSC(&pos[0,0], &number[0,0], &W[0], pos.shape[0], number.shape[0],
             pos.shape[1], BoxSize, threads)
####################################

############## PCS #################
cpdef void PCSc3D(FLOAT[:,::1] pos, FLOAT[:,:,::1] number,
                  FLOAT BoxSize, long threads):
    MASC.PCS(&pos[0,0], &number[0,0,0], NULL, pos.shape[0], number.shape[0],
             pos.shape[1], BoxSize, threads)

cpdef void PCSWc3D(FLOAT[:,::1] pos, FLOAT[:,:,::1] number, FLOAT[::1] W,
                   FLOAT BoxSize, long threads):
    MASC.PCS(&pos[0,0], &number[0,0,0], &W[0], pos.shape[0], number.shape[0],
              pos.shape[1], BoxSize, threads)

cpdef void PCSc2D(FLOAT[:,::1] pos, FLOAT[:,::1] number,
                   FLOAT BoxSize, long threads):
    MASC.PCS(&pos[0,0], &number[0,0], NULL, pos.shape[0], number.shape[0],
              pos.shape[1], BoxSize, threads)
    
cpdef void PCSWc2D(FLOAT[:,::1] pos, FLOAT[:,::1] number, FLOAT[::1] W,
                   FLOAT BoxSize, long threads):
    MASC.PCS(&pos[0,0], &number[0,0], &W[0], pos.shape[0], number.shape[0],
             pos.shape[1], BoxSize, threads)
####################################



# This is the library where we store the value of the different constants
# In order to use this library do:
# import units_library as UL
# U = UL.units()
# kpc = U.kpc_cm

class units:
    def __init__(self):
    
        self.rho_crit = 2.77536627e11 #h^2 Msun/Mpc^3
        self.c_kms    = 3e5           #km/s
        self.Mpc_cm   = 3.0856e24     #cm
        self.kpc_cm   = 3.0856e21     #cm
        self.Msun_g   = 1.989e33      #g
        self.Ymass    = 0.24          #helium mass fraction
        self.mH_g     = 1.6726e-24    #proton mass in grams
        self.yr_s     = 3.15576e7     #seconds
        self.km_cm    = 1e5           #cm
        self.kB       = 1.3806e-26    #gr (km/s)^2 K^{-1}
        self.nu0_MHz  = 1420.0        #21-cm frequency in MHz



import numpy as np
import sys,os,time
import integration_library as IL


elements = 100000
x = np.linspace(0, 10, elements)
y = x**4
steps = elements*100

start = time.time()
I = IL.trapezoidal(x,y)
print('I = %.8f'%I)
print('Time taken = %.2f ms'%(1e3*(time.time()-start)))

start = time.time()
I = IL.simpson(x,y,steps)
print('I = %.8f'%I)
print('Time taken = %.2f ms'%(1e3*(time.time()-start)))



cdef extern from "MAS_c.h":
       ctypedef float FLOAT
       void NGP(FLOAT *pos, FLOAT *number, FLOAT *W, long particles,
                int dims, int axes, FLOAT BoxSize, int threads)
       void CIC(FLOAT *pos, FLOAT *number, FLOAT *W, long particles,
                int dims, int axes, FLOAT BoxSize, int threads)
       void TSC(FLOAT *pos, FLOAT *number, FLOAT *W, long particles,
                int dims, int axes, FLOAT BoxSize, int threads)
       void PCS(FLOAT *pos, FLOAT *number, FLOAT *W, long particles,
                int dims, int axes, FLOAT BoxSize, int threads)



# This script generates the density and velocity fields of a Gaussian density field
# given an input power spectrum. 
import numpy as np
import scipy.fftpack
import density_field_library as DFL
import Pk_library as PKL
import sys

############################### INPUT #########################################
BoxSize = 1000.0  #Mpc/h
dims    = 128     #Generate delta(k) using a grid with dims^3 cells

Rayleigh_sampling = False
precision         = 'double'  #'single' or 'double'

do_sirko   = True   #set delta_k[0,0,0] to 0 (False) or to Pkf[0] (True)
SphereMode = 'cube' #choose between 'sphere' and 'cube'

Pk_file = 'ics_matterpow_10.dat'

seed = 105  #seed for the random numbers

do_lognormal = False

# Pk/Xi parameters
axis    = 0
MAS     = 'None'
threads = 1

z_in    = 10.0    #only needed for the velocities
Omega_m = 0.3175  #only needed for the velocities
Omega_l = 0.6825  #only needed for the velocities

fout_df  = 'Cube_sirko_128^3_z=10.txt'         #file with the density field
f_Pk     = 'Pk_sirko_128^3_density_field.dat'  #file with density field Pk
f_CF     = 'CF_sirko_128^3_density_field.dat'  #file with density field Xi
Pk_theta = 'Pk_theta_field.dat'                #file with the P(k) of \grad V
###############################################################################

dims2, dims3, middle = dims**2, dims**3, dims//2
k_nyquist = dims//2  #Nyquist frequency in units of the fundamental frequency

# compute the value of Omega_m(z), the growth rate and the Hubble function
# Notice that it would be better to compute H(z_in) and the growth rate directly
# from CAMB, but in the past we checked that H_zin*growth_rate is quite 
# insesitive to Omega_radiation
Omega_mz    = Omega_m*(1.0+z_in)**3/(Omega_m*(1.0+z_in)**3+Omega_l)
growth_rate = Omega_mz**0.54545454
H_zin       = 100.0*np.sqrt(Omega_m*(1.0+z_in)**3+Omega_l) #km/s/(Mpc/h)
vel_prefact = -BoxSize/(2.0*np.pi)*H_zin*1.0/(1.0+z_in)*growth_rate #sign is -!!
print('vel_prefact = %.4e km/s'%vel_prefact)
print('a*H(z_i)*f  = %.4e km/s'%(H_zin*1.0/(1.0+z_in)*growth_rate))
# the 1/(BoxSize/(2*pi)) above comes from the ratio \vec{k}/k^2 of \vec{v}_k

# read the P(k) file and set correct units
kf, Pkf = np.loadtxt(Pk_file,unpack=True)
kf = kf*BoxSize/(2.0*np.pi);    Pkf = Pkf/BoxSize**3

# generate the field delta(k)
delta_k,Vx_k,Vy_k,Vz_k = \
    DFL.delta_k(dims, precision, kf, Pkf, Rayleigh_sampling, do_sirko, seed)


############### delta(r) ##################
# compute delta(r) by FFT delta(k)
print('Fourier transforming delta(k)...')
delta_r = scipy.fftpack.fftn(delta_k,overwrite_x=True); del delta_k
delta_r = np.ravel(delta_r)

# since delta(r) has to be real, check that the imaginary part is negligible
ratio = np.absolute(delta_r.imag/delta_r.real)
print('max value of |delta_r.imag/delta_r.real| =',np.max(ratio)); del ratio

# in principle delta(r) will have an imaginary part. Just keep the real one
delta_r = delta_r.real;    print(np.min(delta_r),'< delta(r) <',np.max(delta_r))
print('<delta(r)> =',np.mean(delta_r,dtype=np.float64))
###########################################

################# Vx(r) ###################
# compute Vx(r) by FFT Vx(k)
print('Fourier transforming Vx(k)...')
Vx_r = scipy.fftpack.fftn(Vx_k,overwrite_x=True); del Vx_k
Vx_r = np.ravel(Vx_r); Vx_r*=vel_prefact

#since delta(r) has to be real, check that the imaginary part is negligible
ratio = np.absolute(Vx_r.imag/Vx_r.real)
print('max value of |Vx_r.imag/Vx_r.real| =',np.max(ratio)); del ratio

#in principle Vx(r) will have an imaginary part. Just keep the real one
Vx_r = Vx_r.real
print(np.min(Vx_r),'< Vx <',np.max(Vx_r))
###########################################

################# Vy(r) ###################
# compute Vy(r) by FFT Vy(k)
print('Fourier transforming Vy(k)...')
Vy_r = scipy.fftpack.fftn(Vy_k,overwrite_x=True); del Vy_k
Vy_r = np.ravel(Vy_r); Vy_r*=vel_prefact

# since delta(r) has to be real, check that the imaginary part is negligible
ratio = np.absolute(Vy_r.imag/Vy_r.real)
print('max value of |Vy_r.imag/Vy_r.real| =',np.max(ratio)); del ratio

# in principle Vy(r) will have an imaginary part. Just keep the real one
Vy_r = Vy_r.real
print(np.min(Vy_r),'< Vy <',np.max(Vy_r))
###########################################

################# Vz(r) ###################
# compute Vz(r) by FFT Vz(k)
print('Fourier transforming Vz(k)...')
Vz_r = scipy.fftpack.fftn(Vz_k,overwrite_x=True); del Vz_k
Vz_r = np.ravel(Vz_r); Vz_r*=vel_prefact

# since delta(r) has to be real, check that the imaginary part is negligible
ratio = np.absolute(Vz_r.imag/Vz_r.real)
print('max value of |Vz_r.imag/Vz_r.real| =',np.max(ratio)); del ratio

# in principle Vz(r) will have an imaginary part. Just keep the real one
Vz_r = Vz_r.real
print(np.min(Vz_r),'< Vz <',np.max(Vz_r))
###########################################

# do the lognormal transformation
if do_lognormal:
    sigma_G = np.sqrt(np.mean(delta_r**2,dtype=np.float64))
    print('sigma_G =',sigma_G)
    delta_r = np.exp(delta_r-sigma_G**2/2.0) - 1.0
    print('<delta_NL(r)> =',np.mean(delta_r,dtype=np.float64))
    print(np.min(delta_r),'< delta_NL(r) <',np.max(delta_r))


# save density and velocity field to file
np.save(fout_df, [delta_r, Vx_r, Vy_r, Vz_r])


############ Final checks ###########
# Compute power spectrum & correlation function of the density field 
Pk = PKL.Pk(delta_r, BoxSize, axis, MAS, threads)
CF = PKL.Xi(delta_r, BoxSize, MAS, axis, threads)
np.savetxt(f_Pk, np.transpose([Pk.k3D, Pk.Pk[:,0]]))
np.savetxt(f_CF, np.transpose([CF.r3D, CF.xi[:,0]]));  del delta_r

"""
#compute the power spectrum of the theta=i\grad V
Vx_r = np.reshape(Vx_r,(dims,dims,dims))
Vy_r = np.reshape(Vy_r,(dims,dims,dims))
Vz_r = np.reshape(Vz_r,(dims,dims,dims))

Vx_k = scipy.fftpack.ifftn(Vx_r,overwrite_x=True); del Vx_r
Vy_k = scipy.fftpack.ifftn(Vy_r,overwrite_x=True); del Vy_r
Vz_k = scipy.fftpack.ifftn(Vz_r,overwrite_x=True); del Vz_r

theta_k = DFL.theta(Vx_k,Vy_k,Vz_k,dims)*2.0*np.pi/BoxSize
del Vx_k, Vy_k, Vz_k
theta_r = scipy.fftpack.fftn(theta_k,overwrite_x=True); del theta_k
theta_r = np.ravel(theta_r)
[k,Pk]=PSL.power_spectrum_given_delta(theta_r,dims,BoxSize,
                                      aliasing_method='None')
np.savetxt(Pk_theta,np.transpose([k,Pk])); del theta_r
"""
#####################################






import numpy as np
import scipy.integrate as si
import sys,os,time

#################### FUNCTIONS USED TO COMPUTE INTEGRALS #####################

#we use this function to compute the comoving distance to a given redshift
def func(y,x,Omega_m,Omega_L):
    return [1.0/np.sqrt(Omega_m*(1.0+x)**3+Omega_L)]
##############################################################################



##############################################################################
#This functions computes the comoving distance to redshift z, in Mpc/h
#As input it needs z, Omega_m and Omega_L. It assumes a flat cosmology
def comoving_distance(z,Omega_m,Omega_L):
    H0=100.0 #km/s/(Mpc/h)
    c=3e5    #km/s

    #compute the comoving distance to redshift z
    yinit=[0.0]
    z_limits=[0.0,z]
    I=si.odeint(func,yinit,z_limits,args=(Omega_m,Omega_L),
                rtol=1e-8,atol=1e-8,mxstep=100000,h0=1e-6)[1][0]
    r=c/H0*I

    return r
##############################################################################

##############################################################################
def func_lgf(y,x,Omega_m,Omega_L):
    #print(x, 1.0/(x**3 * (np.sqrt(Omega_m/x**3 + Omega_L))**3))
    return 1.0/(x*np.sqrt(Omega_m/x**3 + Omega_L))**3

#This function computes the linear growth factor. See Eq. 1 of 0006089
#Notice that in that formula H(a) = (Omega_m/a^3+Omega_L)^1/2 and that the 
#growth is D(a), not g(a). We normalize it such as D(a=1)=1
def linear_growth_factor(z,Omega_m,Omega_L):
    
    # compute linear growth factor at z and z=0
    yinit = [0.0];  a_limits = [1e-30, 1.0/(1.0+z), 1.0/(1.0+0.0)]
    I = si.odeint(func_lgf,yinit,a_limits,args=(Omega_m,Omega_L),
                  rtol=1e-10,atol=1e-10,mxstep=100000,h0=1e-20)[1:]
    redshifts = np.array([ [z], [0.0] ])
    Ha = np.sqrt(Omega_m*(1.0+redshifts)**3 + Omega_L)
    D = (5.0*Omega_m/2.0)*Ha*I

    return D[0]/D[1]
##############################################################################


#This function computes the absoption distance:
#dX = H0*(1+z)^2/H(z)*dz
#Omega_m ----> value of the Omega_m cosmological parameter
#Omega_L ----> value of the Omega_L cosmological parameter
#z ----------> cosmological redshift
#BoxSize ----> size of the simulation box in Mpc/h
def absorption_distance(Omega_m,Omega_L,z,BoxSize):
    iterations=40; tol=1e-4; i=0; final=False
    dz_max=10.0; dz_min=0.0; dz=0.5*(dz_min+dz_max)
    r0=comoving_distance(z,Omega_m,Omega_L) #Mpc/h
    while not(final):
        dr=comoving_distance(z+dz,Omega_m,Omega_L)-r0
        if (np.absolute(dr-BoxSize)/BoxSize)<tol or i>iterations:
            final=True
        else:
            i+=1
            if dr>BoxSize:
                dz_max=dz
            else:
                dz_min=dz
            dz=0.5*(dz_min+dz_max)

    dX=(1.0+z)**2/np.sqrt(Omega_m*(1.0+z)**3+Omega_L)*dz
    return dX
##############################################################################

# This routine implements the Takahashi 2012 halofit formula (1208.2701)
# we have neglected the terms with dark energy variation (Eqs. A6 and A7)
# Omega_m ---------------> value of Omega_m at z=0
# Omega_l ---------------> value of Omega_l at z=0
# z ---------------------> redshift
# k_lin,Pk_lin ----------> linear power spectrum at z=0
# for redshifts different to 0 the code computes the growth factor and 
# rescale it. It returns the non-linear P(k) at redshift z. We have checked
# that for a Planck cosmology it reproduces CAMB within 0.6% up to z=5
def Halofit_12(Omega_m,Omega_l,z,k_lin,Pk_lin):

    # compute growth factor at redshift z and rescale P(k)
    Dz = linear_growth_factor(z,Omega_m,Omega_l)
    Pk_lin = Pk_lin*Dz**2
    
    ######### find the value of k_sigma #########
    Rmin,Rmax = 0.01, 10.0  #Mpc/h
    found = False;  precision = 1e-5
    while not(found):
        R = 0.5*(Rmin + Rmax)
        
        yinit = [0.0];  k_limits = [k_lin[0],k_lin[-1]]
        sigma2 = si.odeint(sigma_func, yinit, k_limits, args=(k_lin,Pk_lin,R), 
                           mxstep=1000000, rtol=1e-8, atol=1e-21, 
                           h0=1e-10)[1][0]
        sigma2 = sigma2/(2.0*np.pi**2)

        if abs(sigma2-1.0)<precision:  found = True
        elif sigma2>1.0:  Rmin = R
        else:             Rmax = R
    k_sigma = 1.0/R  #h/Mpc
    #############################################

    ####### compute value of neff and C #########
    # notice that we are doing dlog sigma^2/dlnR = 
    # log(sigma^2(log(R)+log(h))) - log(sigma^2(log(R))) / log(h) = 
    # log(sigma^2(log(R*h))) - log(sigma^2(log(R))) / log(h)
    h = 1.05
    Rp = R*h;  yinit = [0.0];  k_limits = [k_lin[0],k_lin[-1]]
    sigma2p = si.odeint(sigma_func, yinit, k_limits, args=(k_lin,Pk_lin,Rp), 
                        mxstep=1000000, rtol=1e-8, atol=1e-21, h0=1e-10)[1][0]
    sigma2p = sigma2p/(2.0*np.pi**2)

    Rm = R/h;  yinit = [0.0];  k_limits = [k_lin[0],k_lin[-1]]
    sigma2m = si.odeint(sigma_func, yinit, k_limits, args=(k_lin,Pk_lin,Rm), 
                        mxstep=1000000, rtol=1e-8, atol=1e-21, h0=1e-10)[1][0]
    sigma2m = sigma2m/(2.0*np.pi**2)

    # sanity check
    if abs(sigma2p-sigma2)<1e3*precision or abs(sigma2-sigma2m)<1e3*precision:
        print('value of h too small for given precision');  sys.exit()

    neff = -(np.log(sigma2p)-np.log(sigma2m))/(2.0*np.log(h)) - 3.0
    C = -(np.log(sigma2p) - 2.0*np.log(sigma2) + np.log(sigma2m))/np.log(h)**2
    #############################################

    ################ constants ##################
    Omegamz = Omega_m*(1.0+z)**3/(Omega_m*(1.0+z)**3 + Omega_l)

    f1 = Omegamz**(-0.0307);  f2 = Omegamz**(-0.0585);  f3 = Omegamz**(0.0743)

    an = 10**(1.5222 + 2.8553*neff + 2.3706*neff**2 + 0.9903*neff**3 + \
                  0.2250*neff**4 - 0.6038*C)
    bn = 10**(-0.5642 + 0.5864*neff + 0.5716*neff**2 - 1.5474*C)
    cn = 10**(0.3698 + 2.0404*neff + 0.8161*neff**2 + 0.5869*C)
    gamman = 0.1971 - 0.0843*neff + 0.8460*C
    alphan = abs(6.0835 + 1.3373*neff - 0.1959*neff**2 - 5.5274*C)
    betan  = 2.0379 - 0.7354*neff + 0.3157*neff**2 + 1.2490*neff**3 + \
        0.3980*neff**4 - 0.1682*C
    mun = 0.0
    nun = 10**(5.2105 + 3.6902*neff)
    ############################################


    Pk_hf  = np.zeros(len(k_lin),dtype=np.float64)
    for i,k in enumerate(k_lin):

        # dimensionless linear power spectrum
        delta2_lin = k**3*Pk_lin[i]/(2.0*np.pi**2)

        y = (k/k_sigma);  fy = y/4.0 + y**2/8.0

        # two-halo term
        delta2_Q = delta2_lin*(1.0 + delta2_lin)**betan/\
            (1.0 + alphan*delta2_lin)*np.exp(-fy)

        # one-halo term
        delta2_HH = an*y**(3.0*f1)/(1.0 + bn*y**f2 + (cn*f3*y)**(3.0-gamman))
        delta2_H  = delta2_HH/(1.0 + mun/y +  nun/y**2)

        # total non-linear dimensionless power spectrum
        delta2_hf = delta2_Q + delta2_H

        # non-linear power spectrum
        Pk_hf[i]  = (2*np.pi**2)*delta2_hf/k**3

    return Pk_hf
    #############################################

def sigma_func(y,x,k,Pk,R):
    return [np.interp(x,k,Pk)*x**2*np.exp(-x**2*R**2)]
##############################################################################







###############################################################################
################################### USAGE #####################################
###############################################################################

###### comoving distance ######
"""
z=3.0
Omega_m=0.3
Omega_L=0.7

r=comoving_distance(z,Omega_m,Omega_L)
print('comoving distance to z = %2.2f ---> %f Mpc/h'%(z,r))
"""

###### linear growth factor ######
"""
z       = 1.0
Omega_m = 0.308
Omega_l = 0.692
h       = 0.6781

Da = linear_growth_factor(z,Omega_m,Omega_l,h)
print('Linear growth factor at z = %.1f : %.3e'%(z,Da))
"""

###### absorption distance ######
"""
Omega_m=0.274247
Omega_L=0.725753
z=3.0
BoxSize=60.0 #Mpc/h

dX=absorption_distance(Omega_m,Omega_L,z,BoxSize)
print('dX =',dX)
"""

###### halofit P(k) ######
"""
Omega_m = 0.3175
Omega_l = 0.6825
z       = 1.0
k_lin,Pk_lin = np.loadtxt('ics_matterpow_0.dat',unpack=True)
Pk_hf = Halofit_12(Omega_m,Omega_l,z,k_lin,Pk_lin)
"""



import numpy as np 
cimport numpy as np
import time,sys,os
cimport cython
cimport integration as CI
from libc.math cimport sqrt,pow,sin,cos,floor,fabs,log10

############### example 1 ################
cdef void example(double x, double y[], double dydx[],
                   double a[], double b[], long elements):
    dydx[0] = x*x + 2.0*x + 3.0
##########################################

############### example 2 ################
cdef void example2(double x, double y[], double dydx[],
                   double a[], double b[], long elements):

    cdef int index
    cdef double a_interp

    if x<=a[0]:             a_interp = b[0]
    elif x>=a[elements-1]:  a_interp = b[elements-1]
    else:
        index = <int>((x-a[0])/(a[elements-1]-a[0])*(elements-1))
        a_interp = (b[index+1] - b[index])/(a[index+1]-a[index])*(x-a[index]) + b[index]
    dydx[0] = a_interp
##########################################

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True) 
cdef void linear(double x, double y[], double dydx[],
                 double a[], double b[], long elements):

    cdef int index
    cdef double b_interp

    if x<=a[0]:             b_interp = b[0]
    elif x>=a[elements-1]:  b_interp = b[elements-1]
    else:
        index = <int>((x-a[0])/(a[elements-1]-a[0])*(elements-1))
        b_interp = (b[index+1] - b[index])/(a[index+1]-a[index])*(x-a[index]) + b[index]
    dydx[0] = b_interp

    
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True) 
cdef void sigma(double x, double y[], double dydx[],
                double a[], double b[], long elements):

    # a is the input log10(k)
    # b is the input log10(Pk)
    # a is equally spaced in log10
    cdef int index
    cdef double Pk_interp, logx

    logx = log10(x)
    
    if logx<=a[0]:             Pk_interp = b[0]
    elif logx>=a[elements-1]:  Pk_interp = b[elements-1]
    else:
        index = <int>((logx-a[0])/(a[elements-1]-a[0])*(elements-1))
        Pk_interp = (b[index+1] - b[index])/(a[index+1]-a[index])*(logx-a[index]) + b[index]
    
    dydx[0] = Pk_interp
    #dydx[0] = 10**(Pk_interp)  #old method

#######################################################################
############### Trapezoidal rule ##############
cpdef double trapezoidal(double[::1] x_array, double[::1] y_array):
    return CI.rectangular(x_array[0], x_array[-1], &x_array[0], &y_array[0],
                          x_array.shape[0])
###############################################

############### Simpson rule ##################
cpdef double simpson(double[::1] x_array, double[::1] y_array, long steps):
    return CI.simpson(x_array[0], x_array[-1], &x_array[0], &y_array[0],
                      x_array.shape[0], steps)
###############################################

# RK4 example1 for debugging
cpdef RK4_example(double[::1] yinit, double x1, double x2, long nstep):

    cdef double *result      
    result =  CI.rkdumb(&yinit[0], yinit.shape[0], x1, x2, nstep,
                        NULL, NULL, 0, example)
    # cast pointer to cython memory view. Then memory view to numpy array
    return np.asarray(<double[:yinit.shape[0]]> result)

# RK4 example2 for debugging
cpdef RK4_example2(double[::1] yinit, double x1, double x2, long nstep,
                   double[::1] a, double[::1] b):

    cdef double *result      
    result =  CI.rkdumb(&yinit[0], yinit.shape[0], x1, x2, nstep,
                        &a[0], &b[0], a.shape[0], example2)
    # cast pointer to cython memory view. Then memory view to numpy array
    return np.asarray(<double[:yinit.shape[0]]> result)

    
# RK5 with adaptive step example1
cpdef odeint_example1(double[::1] yinit, double x1, double x2, double eps,
                      double h1, double hmin, verbose=True):

    cdef long nok, nbad
    nok = nbad = 0
    
    CI.odeint(&yinit[0], yinit.shape[0], x1, x2, eps, h1, hmin, &nok, &nbad,
              NULL, NULL, 0, example)

    if verbose:
        print('Total steps = %ld'%(nok+nbad))
        print('OK    steps = %ld'%nok)
        print('BAD   steps = %ld'%nbad)

    return np.asarray(yinit)

# RK5 with adaptive step example2
cpdef odeint_example2(double[::1] yinit, double x1, double x2, double eps,
                      double h1, double hmin, double[::1] a, double[::1] b,
                      verbose=True):

    cdef long nok, nbad
    nok = nbad = 0
    
    CI.odeint(&yinit[0], yinit.shape[0], x1, x2, eps, h1, hmin, &nok, &nbad,
              &a[0], &b[0], a.shape[0], example2)

    if verbose:
        print('Total steps = %ld'%(nok+nbad))
        print('OK    steps = %ld'%nok)
        print('BAD   steps = %ld'%nbad)

    return np.asarray(yinit)

# RK5 with adaptive step
cpdef odeint(double[::1] yinit, double x1, double x2, double eps,
             double h1, double hmin, double[::1] a, double[::1] b,
             function, verbose=True):

    cdef long nok, nbad
    nok = nbad = 0

    if function=='log' or function=='sigma':
        CI.odeint(&yinit[0], yinit.shape[0], x1, x2, eps, h1, hmin, &nok, &nbad,
                  &a[0], &b[0], b.shape[0], sigma)

    elif function=='linear':
        CI.odeint(&yinit[0], yinit.shape[0], x1, x2, eps, h1, hmin, &nok, &nbad,
                  &a[0], &b[0], b.shape[0], linear)

    else:  raise Exception('Incorrect function!')
        

    if verbose:
        print('Total steps = %ld'%(nok+nbad))
        print('OK    steps = %ld'%nok)
        print('BAD   steps = %ld'%nbad)

    return np.asarray(yinit)
